from unittest.mock import Mock

from .. import test
from ...acl import useracl
from ...acl.test import patch_user_acl
from ...categories.models import Category
from ...conf.test import override_dynamic_settings
from ...conftest import get_cache_versions
from ...users.test import AuthenticatedUserTestCase
from ..checksums import update_post_checksum
from ..events import record_event
from ..moderation import hide_post
from ..moderation import threads as threads_moderation

cache_versions = get_cache_versions()


def patch_category_acl(new_acl=None):
    def patch_acl(_, user_acl):
        category = Category.objects.get(slug="first-category")
        category_acl = user_acl["categories"][category.id]

        # reset category ACL to single predictable state
        category_acl.update(
            {
                "can_see": 1,
                "can_browse": 1,
                "can_see_all_threads": 1,
                "can_see_own_threads": 0,
                "can_hide_threads": 0,
                "can_approve_content": 0,
                "can_edit_posts": 0,
                "can_hide_posts": 0,
                "can_hide_own_posts": 0,
                "can_close_threads": 0,
                "post_edit_time": 0,
                "can_hide_events": 0,
            }
        )

        if new_acl:
            category_acl.update(new_acl)

    return patch_user_acl(patch_acl)


class ThreadViewTestCase(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()

        self.category = Category.objects.get(slug="first-category")
        self.thread = test.post_thread(category=self.category)


class ThreadVisibilityTests(ThreadViewTestCase):
    def test_thread_displays(self):
        """thread view has no showstoppers"""
        response = self.client.get(self.thread.get_absolute_url())
        self.assertContains(response, self.thread.title)

    def test_view_shows_owner_thread(self):
        """view handles "owned threads" only"""
        with patch_category_acl({"can_see_all_threads": 0}):
            response = self.client.get(self.thread.get_absolute_url())
            self.assertEqual(response.status_code, 404)

            self.thread.starter = self.user
            self.thread.save()

            response = self.client.get(self.thread.get_absolute_url())
            self.assertContains(response, self.thread.title)

    def test_view_validates_category_permissions(self):
        """view validates category visiblity"""
        with patch_category_acl({"can_see": 0}):
            response = self.client.get(self.thread.get_absolute_url())
            self.assertEqual(response.status_code, 404)

        with patch_category_acl({"can_browse": 0}):
            response = self.client.get(self.thread.get_absolute_url())
            self.assertEqual(response.status_code, 404)

    def test_view_shows_unapproved_thread(self):
        """view handles unapproved thread"""
        with patch_category_acl({"can_approve_content": 0}):
            self.thread.is_unapproved = True
            self.thread.save()

            response = self.client.get(self.thread.get_absolute_url())
            self.assertEqual(response.status_code, 404)

        # grant permission to see unapproved content
        with patch_category_acl({"can_approve_content": 1}):
            response = self.client.get(self.thread.get_absolute_url())
            self.assertContains(response, self.thread.title)

            # make test user thread's owner and remove permission to see unapproved
            # user should be able to see thread as its author anyway
            self.thread.starter = self.user
            self.thread.save()

        with patch_category_acl({"can_approve_content": 0}):
            response = self.client.get(self.thread.get_absolute_url())
            self.assertContains(response, self.thread.title)

    def test_view_shows_hidden_thread(self):
        """view handles hidden thread"""
        with patch_category_acl({"can_hide_threads": 0}):
            self.thread.is_hidden = True
            self.thread.save()

            response = self.client.get(self.thread.get_absolute_url())
            self.assertEqual(response.status_code, 404)

        # threads owners are not extempt from hidden threads check
        self.thread.starter = self.user
        self.thread.save()

        response = self.client.get(self.thread.get_absolute_url())
        self.assertEqual(response.status_code, 404)

        # grant permission to see hidden content
        with patch_category_acl({"can_hide_threads": 1}):
            response = self.client.get(self.thread.get_absolute_url())
            self.assertContains(response, self.thread.title)


class ThreadPostsVisibilityTests(ThreadViewTestCase):
    def test_post_renders(self):
        """post renders"""
        post = test.reply_thread(self.thread, poster=self.user)

        response = self.client.get(self.thread.get_absolute_url())
        self.assertContains(response, post.get_absolute_url())

    def test_invalid_post_renders(self):
        """invalid post renders"""
        post = test.reply_thread(self.thread, poster=self.user)

        post.parsed = "fiddled post content"
        post.save()

        response = self.client.get(self.thread.get_absolute_url())
        self.assertContains(response, post.get_absolute_url())
        self.assertContains(response, "This post's contents cannot be displayed.")
        self.assertNotContains(response, post.parsed)

    def test_hidden_post_visibility(self):
        """hidden post renders correctly"""
        post = test.reply_thread(self.thread, message="Hello, I'm hidden post!")
        hide_post(self.user, post)

        response = self.client.get(self.thread.get_absolute_url())
        self.assertContains(response, post.get_absolute_url())
        self.assertContains(
            response, "This post is hidden. You cannot not see its contents."
        )
        self.assertNotContains(response, post.parsed)

        # posts authors are not extempt from seeing hidden posts content
        post.posted_by = self.user
        post.save()

        response = self.client.get(self.thread.get_absolute_url())
        self.assertContains(response, post.get_absolute_url())
        self.assertContains(
            response, "This post is hidden. You cannot not see its contents."
        )
        self.assertNotContains(response, post.parsed)

        # permission to hide own posts isn't enought to see post content
        with patch_category_acl({"can_hide_own_posts": 1}):
            response = self.client.get(self.thread.get_absolute_url())
            self.assertContains(response, post.get_absolute_url())
            self.assertContains(
                response, "This post is hidden. You cannot not see its contents."
            )
            self.assertNotContains(response, post.parsed)

        # post's content is displayed after permission to see posts is granted
        with patch_category_acl({"can_hide_posts": 1}):
            response = self.client.get(self.thread.get_absolute_url())
            self.assertContains(response, post.get_absolute_url())
            self.assertContains(
                response,
                "This post is hidden. Only users with permission may see its contents.",
            )
            self.assertNotContains(
                response, "This post is hidden. You cannot not see its contents."
            )
            self.assertContains(response, post.parsed)

    def test_unapproved_post_visibility(self):
        """
        unapproved post renders for its author and users with perm to approve content
        """
        post = test.reply_thread(self.thread, is_unapproved=True)

        # post is hdden because we aren't its author nor user with permission to approve
        response = self.client.get(self.thread.get_absolute_url())
        self.assertNotContains(response, post.get_absolute_url())

        # post displays because we have permission to approve unapproved content
        with patch_category_acl({"can_approve_content": 1}):
            response = self.client.get(self.thread.get_absolute_url())
            self.assertContains(response, post.get_absolute_url())
            self.assertContains(response, "This post is unapproved.")
            self.assertContains(response, post.parsed)

        # post displays because we are its author
        with patch_category_acl({"can_approve_content": 0}):
            post.poster = self.user
            post.save()

            response = self.client.get(self.thread.get_absolute_url())
            self.assertContains(response, post.get_absolute_url())
            self.assertContains(response, "This post is unapproved.")
            self.assertContains(response, post.parsed)


class ThreadEventVisibilityTests(ThreadViewTestCase):
    def test_thread_events_render(self):
        """different thread events render"""
        TEST_ACTIONS = [
            (
                threads_moderation.pin_thread_globally,
                "Thread has been pinned globally.",
            ),
            (threads_moderation.pin_thread_locally, "Thread has been pinned locally."),
            (threads_moderation.unpin_thread, "Thread has been unpinned."),
            (threads_moderation.approve_thread, "Thread has been approved."),
            (threads_moderation.close_thread, "Thread has been closed."),
            (threads_moderation.open_thread, "Thread has been opened."),
            (threads_moderation.hide_thread, "Thread has been made hidden."),
            (threads_moderation.unhide_thread, "Thread has been revealed."),
        ]

        self.thread.is_unapproved = True
        self.thread.save()

        for action, message in TEST_ACTIONS:
            self.thread.post_set.filter(is_event=True).delete()

            with patch_category_acl({"can_approve_content": 1, "can_hide_threads": 1}):
                user_acl = useracl.get_user_acl(self.user, cache_versions)
                request = Mock(user=self.user, user_acl=user_acl, user_ip="127.0.0.1")
                action(request, self.thread)

                event = self.thread.post_set.filter(is_event=True)[0]

                # event renders
                response = self.client.get(self.thread.get_absolute_url())
                self.assertContains(response, event.get_absolute_url())
                self.assertContains(response, message)

            # hidden events don't render without permission
            with patch_category_acl({"can_approve_content": 1, "can_hide_threads": 1}):
                hide_post(self.user, event)
                response = self.client.get(self.thread.get_absolute_url())
                self.assertNotContains(response, event.get_absolute_url())
                self.assertNotContains(response, message)

            # hidden event renders with permission
            with patch_category_acl(
                {"can_approve_content": 1, "can_hide_threads": 1, "can_hide_events": 1}
            ):
                hide_post(self.user, event)
                response = self.client.get(self.thread.get_absolute_url())
                self.assertContains(response, event.get_absolute_url())
                self.assertContains(response, message)
                self.assertContains(response, "Hidden by")

            # Event is only loaded if thread has events flag
            with patch_category_acl(
                {"can_approve_content": 1, "can_hide_threads": 1, "can_hide_events": 1}
            ):
                self.thread.has_events = False
                self.thread.save()

                response = self.client.get(self.thread.get_absolute_url())
                self.assertNotContains(response, event.get_absolute_url())

    @override_dynamic_settings(events_per_page=4)
    def test_events_limit(self):
        """forum will trim oldest events if theres more than allowed by config"""
        events = []
        for _ in range(5):
            request = Mock(user=self.user, user_ip="127.0.0.1")
            event = record_event(request, self.thread, "closed")
            events.append(event)

        # test that only events within limits were rendered
        response = self.client.get(self.thread.get_absolute_url())

        for event in events[4:]:
            self.assertContains(response, event.get_absolute_url())
        self.assertNotContains(response, events[0].get_absolute_url())

    @override_dynamic_settings(posts_per_page=10, events_per_page=4)
    def test_events_dont_take_space(self):
        """events dont take space away from posts"""
        posts_limit = 10
        events_limit = 4
        events = []

        for _ in range(events_limit + 5):
            request = Mock(user=self.user, user_ip="127.0.0.1")
            event = record_event(request, self.thread, "closed")
            events.append(event)

        posts = []
        for _ in range(posts_limit - 1):
            post = test.reply_thread(self.thread)
            posts.append(post)

        # test that all events and posts within limits were rendered
        response = self.client.get(self.thread.get_absolute_url())

        for event in events[5:]:
            self.assertContains(response, event.get_absolute_url())
        for post in posts:
            self.assertContains(response, post.get_absolute_url())

        # add second page to thread with more events
        for _ in range(posts_limit):
            post = test.reply_thread(self.thread)
        for _ in range(events_limit):
            request = Mock(user=self.user, user_ip="127.0.0.1")
            event = record_event(request, self.thread, "closed")
            events.append(event)

        # see first page
        response = self.client.get(self.thread.get_absolute_url())

        for event in events[5:events_limit]:
            self.assertContains(response, event.get_absolute_url())
        for post in posts[: posts_limit - 1]:
            self.assertContains(response, post.get_absolute_url())

        # see second page
        response = self.client.get("%s2/" % self.thread.get_absolute_url())
        for event in events[5 + events_limit :]:
            self.assertContains(response, event.get_absolute_url())
        for post in posts[posts_limit - 1 :]:
            self.assertContains(response, post.get_absolute_url())

    def test_changed_thread_title_event_renders(self):
        """changed thread title event renders"""
        request = Mock(user=self.user, user_ip="127.0.0.1")
        threads_moderation.change_thread_title(
            request, self.thread, "Lorem renamed ipsum!"
        )

        event = self.thread.post_set.filter(is_event=True)[0]
        self.assertEqual(event.event_type, "changed_title")

        # event renders
        response = self.client.get(self.thread.get_absolute_url())
        self.assertContains(response, event.get_absolute_url())
        self.assertContains(response, "title has been changed from")
        self.assertContains(response, self.thread.title)

    def test_thread_move_event_renders(self):
        """moved thread event renders"""
        self.thread.category = self.thread.category.parent
        self.thread.save()

        request = Mock(user=self.user, user_ip="127.0.0.1")
        threads_moderation.move_thread(request, self.thread, self.category)

        event = self.thread.post_set.filter(is_event=True)[0]
        self.assertEqual(event.event_type, "moved")

        # event renders
        response = self.client.get(self.thread.get_absolute_url())
        self.assertContains(response, event.get_absolute_url())
        self.assertContains(response, "Thread has been moved from")

    def test_thread_merged_event_renders(self):
        """merged thread event renders"""
        request = Mock(user=self.user, user_ip="127.0.0.1")
        other_thread = test.post_thread(category=self.category)
        threads_moderation.merge_thread(request, self.thread, other_thread)

        event = self.thread.post_set.filter(is_event=True)[0]
        self.assertEqual(event.event_type, "merged")

        # event renders
        response = self.client.get(self.thread.get_absolute_url())
        self.assertContains(response, event.get_absolute_url())
        self.assertContains(response, "thread has been merged into this thread")


class ThreadAttachmentsViewTests(ThreadViewTestCase):
    def mock_attachment_cache(self, data):
        json = {
            "url": {},
            "size": 16914,
            "filename": "Archiwum.zip",
            "filetype": "ZIP",
            "is_image": False,
            "uploaded_on": "2016-10-22T21:17:40.408710Z",
            "uploader_name": "User",
        }

        json.update(data)
        return json

    def test_attachments_display(self):
        """thread posts show list of attachments below them"""
        post = self.thread.first_post

        post.attachments_cache = [
            self.mock_attachment_cache(
                {
                    "url": {
                        "index": "/attachment/loremipsum-123/",
                        "thumb": None,
                        "uploader": "/user/user-123/",
                    },
                    "filename": "Archiwum-1.zip",
                }
            ),
            self.mock_attachment_cache(
                {
                    "url": {
                        "index": "/attachment/loremipsum-223/",
                        "thumb": "/attachment/thumb/loremipsum-223/",
                        "uploader": "/user/user-223/",
                    },
                    "is_image": True,
                    "filename": "Archiwum-2.zip",
                }
            ),
            self.mock_attachment_cache(
                {
                    "url": {
                        "index": "/attachment/loremipsum-323/",
                        "thumb": None,
                        "uploader": "/user/user-323/",
                    },
                    "filename": "Archiwum-3.zip",
                }
            ),
        ]
        post.save()

        # attachments render
        response = self.client.get(self.thread.get_absolute_url())

        for attachment in post.attachments_cache:
            self.assertContains(response, attachment["filename"])
            self.assertContains(response, attachment["uploader_name"])
            self.assertContains(response, attachment["url"]["index"])
            self.assertContains(response, attachment["url"]["uploader"])

            if attachment["url"]["thumb"]:
                self.assertContains(response, attachment["url"]["thumb"])


class ThreadPollViewTests(ThreadViewTestCase):
    def test_poll_voted_display(self):
        """view has no showstoppers when displaying voted poll"""
        poll = test.post_poll(self.thread, self.user)

        response = self.client.get(self.thread.get_absolute_url())
        self.assertContains(response, poll.question)
        self.assertContains(response, "4 votes")
        self.assertNotContains(response, "Save your vote")

    def test_poll_unvoted_display(self):
        """view has no showstoppers when displaying poll vote form"""
        poll = test.post_poll(self.thread, self.user)
        poll.pollvote_set.all().delete()

        response = self.client.get(self.thread.get_absolute_url())
        self.assertContains(response, poll.question)
        self.assertContains(response, "Save your vote")

    def test_poll_anonymous_view(self):
        """view has no showstoppers when displaying poll to anon user"""
        poll = test.post_poll(self.thread, self.user)

        self.logout_user()

        response = self.client.get(self.thread.get_absolute_url())
        self.assertContains(response, poll.question)
        self.assertContains(response, "4 votes")
        self.assertNotContains(response, "Save your vote")


class ThreadLikedPostsViewTests(ThreadViewTestCase):
    def test_liked_posts_display(self):
        """view has no showstoppers on displaying posts with likes"""
        test.like_post(self.thread.first_post, self.user)

        response = self.client.get(self.thread.get_absolute_url())
        self.assertContains(response, '"is_liked": true')

    def test_liked_posts_no_permission(self):
        """
        view has no showstoppers on displaying posts with likes without perm
        """
        test.like_post(self.thread.first_post, self.user)

        with patch_category_acl({"can_see_posts_likes": 0}):
            response = self.client.get(self.thread.get_absolute_url())
            self.assertNotContains(response, '"is_liked": true')
            self.assertNotContains(response, '"is_liked": false')
            self.assertContains(response, '"is_liked": null')


class ThreadAnonViewTests(ThreadViewTestCase):
    def test_anonymous_user_view_no_showstoppers_display(self):
        """kitchensink thread view has no showstoppers for anons"""
        request = Mock(user=self.user, user_ip="127.0.0.1")

        poll = test.post_poll(self.thread, self.user)
        event = record_event(request, self.thread, "closed")

        hidden_event = record_event(request, self.thread, "opened")
        hide_post(self.user, hidden_event)

        unapproved_post = test.reply_thread(self.thread, is_unapproved=True)
        post = test.reply_thread(self.thread)

        self.logout_user()

        response = self.client.get(self.thread.get_absolute_url())
        self.assertContains(response, poll.question)
        self.assertContains(response, event.get_absolute_url())
        self.assertContains(response, post.get_absolute_url())
        self.assertNotContains(response, hidden_event.get_absolute_url())
        self.assertNotContains(response, unapproved_post.get_absolute_url())


class ThreadUnicodeSupportTests(ThreadViewTestCase):
    def test_category_name(self):
        """unicode in category name causes no showstopper"""
        self.category.name = "Łódź"
        self.category.save()

        with patch_category_acl():
            response = self.client.get(self.thread.get_absolute_url())
            self.assertEqual(response.status_code, 200)

    def test_thread_title(self):
        """unicode in thread title causes no showstopper"""
        self.thread.title = "Łódź"
        self.thread.slug = "Lodz"
        self.thread.save()

        with patch_category_acl():
            response = self.client.get(self.thread.get_absolute_url())
            self.assertEqual(response.status_code, 200)

    def test_post_content(self):
        """unicode in thread title causes no showstopper"""
        self.thread.first_post.original = "Łódź"
        self.thread.first_post.parsed = "<p>Łódź</p>"

        update_post_checksum(self.thread.first_post)
        self.thread.first_post.save()

        with patch_category_acl():
            response = self.client.get(self.thread.get_absolute_url())
            self.assertEqual(response.status_code, 200)

    def test_user_rank(self):
        """unicode in user rank causes no showstopper"""
        self.user.title = "Łódź"
        self.user.rank.name = "Łódź"
        self.user.rank.title = "Łódź"
        self.user.rank.save()
        self.user.save()

        with patch_category_acl():
            response = self.client.get(self.thread.get_absolute_url())
            self.assertEqual(response.status_code, 200)



from django import template

register = template.Library()


@register.filter
def isdescriptionshort(string):
    string_lowered = string.lower()
    return string_lowered.count("<p") == 1 and not string_lowered.count("<br")



from django.contrib.auth import get_user_model
from django.http import Http404
from django.shortcuts import get_object_or_404, redirect, render
from django.views import View

from ...acl.objectacl import add_acl_to_obj
from ...core.shortcuts import paginate, pagination_dict, validate_slug
from ..bans import get_user_ban
from ..online.utils import get_user_status
from ..pages import user_profile
from ..profilefields import profilefields, serialize_profilefields_data
from ..serializers import BanDetailsSerializer, UsernameChangeSerializer, UserSerializer
from ..viewmodels import Followers, Follows, UserPosts, UserThreads

User = get_user_model()


class ProfileView(View):
    def get(self, request, *args, **kwargs):
        profile = self.get_profile(request, kwargs.pop("pk"), kwargs.pop("slug"))

        # resolve that we can display requested section
        sections = user_profile.get_sections(request, profile)
        active_section = self.get_active_section(sections)

        if not active_section:
            raise Http404()

        profile.status = get_user_status(request, profile)
        context_data = self.get_context_data(request, profile)

        self.complete_frontend_context(request, profile, sections)
        self.complete_context_data(request, profile, sections, context_data)

        return render(request, self.template_name, context_data)

    def get_profile(self, request, pk, slug):
        queryset = User.objects.select_related("rank", "online_tracker", "ban_cache")

        profile = get_object_or_404(queryset, pk=pk)

        if not profile.is_active and not request.user.is_staff:
            raise Http404()

        validate_slug(profile, slug)
        add_acl_to_obj(request.user_acl, profile)

        return profile

    def get_active_section(self, sections):
        for section in sections:
            if section["is_active"]:
                return section

    def get_context_data(self, request, profile):
        return {}

    def complete_frontend_context(self, request, profile, sections):
        request.frontend_context["PROFILE_PAGES"] = []
        for section in sections:
            request.frontend_context["PROFILE_PAGES"].append(
                {
                    "name": str(section["name"]),
                    "icon": section["icon"],
                    "meta": section.get("metadata"),
                    "component": section["component"],
                }
            )

        request.frontend_context["PROFILE"] = UserProfileSerializer(
            profile, context={"request": request}
        ).data

        if not profile.is_active:
            request.frontend_context["PROFILE"]["is_active"] = False
        if profile.is_deleting_account:
            request.frontend_context["PROFILE"]["is_deleting_account"] = True

    def complete_context_data(self, request, profile, sections, context):
        context["profile"] = profile

        context["sections"] = sections
        for section in sections:
            if section["is_active"]:
                context["active_section"] = section
                break

        if request.user.is_authenticated:
            is_authenticated_user = profile.pk == request.user.pk
            context.update(
                {
                    "is_authenticated_user": is_authenticated_user,
                    "show_email": is_authenticated_user,
                }
            )

            if not context["show_email"]:
                context["show_email"] = request.user_acl["can_see_users_emails"]
        else:
            context.update({"is_authenticated_user": False, "show_email": False})


class LandingView(ProfileView):
    def get(self, request, *args, **kwargs):
        profile = self.get_profile(request, kwargs.pop("pk"), kwargs.pop("slug"))

        return redirect(
            user_profile.get_default_link(), slug=profile.slug, pk=profile.pk
        )


class UserPostsView(ProfileView):
    template_name = "misago/profile/posts.html"

    def get_context_data(self, request, profile):
        feed = UserPosts(request, profile)

        request.frontend_context["POSTS"] = feed.get_frontend_context()
        return feed.get_template_context()


class UserThreadsView(ProfileView):
    template_name = "misago/profile/threads.html"

    def get_context_data(self, request, profile):
        feed = UserThreads(request, profile)

        request.frontend_context["POSTS"] = feed.get_frontend_context()
        return feed.get_template_context()


class UserFollowersView(ProfileView):
    template_name = "misago/profile/followers.html"

    def get_context_data(self, request, profile):
        users = Followers(request, profile)

        request.frontend_context["PROFILE_FOLLOWERS"] = users.get_frontend_context()
        return users.get_template_context()


class UserFollowsView(ProfileView):
    template_name = "misago/profile/follows.html"

    def get_context_data(self, request, profile):
        users = Follows(request, profile)

        request.frontend_context["PROFILE_FOLLOWS"] = users.get_frontend_context()
        return users.get_template_context()


class UserProfileDetailsView(ProfileView):
    template_name = "misago/profile/details.html"

    def get_context_data(self, request, profile):
        details = serialize_profilefields_data(request, profilefields, profile)

        request.frontend_context["PROFILE_DETAILS"] = details

        return {"profile_details": details}


class UserUsernameHistoryView(ProfileView):
    template_name = "misago/profile/username_history.html"

    def get_context_data(self, request, profile):
        queryset = profile.namechanges.select_related("user", "changed_by")
        queryset = queryset.order_by("-id")

        page = paginate(queryset, None, 14, 4)

        data = pagination_dict(page)
        data.update(
            {"results": UsernameChangeSerializer(page.object_list, many=True).data}
        )

        request.frontend_context["PROFILE_NAME_HISTORY"] = data

        return {"history": page.object_list, "count": data["count"]}


class UserBanView(ProfileView):
    template_name = "misago/profile/ban_details.html"

    def get_context_data(self, request, profile):
        ban = get_user_ban(profile, request.cache_versions)

        request.frontend_context["PROFILE_BAN"] = BanDetailsSerializer(ban).data

        return {"ban": ban}


UserProfileSerializer = UserSerializer.subset_fields(
    "id",
    "username",
    "slug",
    "email",
    "joined_on",
    "rank",
    "title",
    "avatars",
    "is_avatar_locked",
    "signature",
    "is_signature_locked",
    "followers",
    "following",
    "threads",
    "posts",
    "acl",
    "is_followed",
    "is_blocked",
    "real_name",
    "status",
    "api",
    "url",
)



# Generated by Django 1.11.1 on 2017-05-21 17:52
import django.contrib.postgres.indexes
from django.contrib.postgres.operations import BtreeGinExtension
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("misago_threads", "0004_update_settings")]

    operations = [
        BtreeGinExtension(),
        migrations.AddIndex(
            model_name="post",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["search_vector"], name="misago_thre_search__b472a2_gin"
            ),
        ),
    ]



from . import exceptionhandler
from .utils import is_request_to_misago


class ExceptionHandlerMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        return self.get_response(request)

    def process_exception(self, request, exception):
        request_is_to_misago = is_request_to_misago(request)
        misago_can_handle_exception = exceptionhandler.is_misago_exception(exception)

        if request_is_to_misago and misago_can_handle_exception:
            return exceptionhandler.handle_misago_exception(request, exception)


class FrontendContextMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        request.include_frontend_context = True
        request.frontend_context = {}
        return self.get_response(request)



import pytest
from django.urls import reverse

from ....cache.test import assert_invalidates_cache
from ....test import assert_contains, assert_has_error_message
from ... import THEME_CACHE


@pytest.fixture
def create_link(theme):
    return reverse("misago:admin:themes:new-css-file", kwargs={"pk": theme.pk})


@pytest.fixture
def edit_link(theme, css):
    return reverse(
        "misago:admin:themes:edit-css-file", kwargs={"pk": theme.pk, "css_pk": css.pk}
    )


@pytest.fixture
def data():
    return {"name": "test.css", "source": ".page-header { padding: 0}"}


def test_css_creation_form_is_displayed(admin_client, create_link):
    response = admin_client.get(create_link)
    assert response.status_code == 200
    assert_contains(response, "New CSS")


def test_css_can_be_created(theme, admin_client, create_link, data):
    admin_client.post(create_link, data)
    assert theme.css.exists()


def test_css_is_created_with_entered_name(theme, admin_client, create_link, data):
    admin_client.post(create_link, data)
    assert theme.css.last().name == data["name"]


def test_created_source_file_contains_css_entered_by_user(
    theme, admin_client, create_link, data
):
    admin_client.post(create_link, data)
    css = theme.css.last()
    assert css.source_file.read().decode("utf-8") == data["source"]


def test_source_file_is_created_in_theme_directory(
    theme, admin_client, create_link, data
):
    admin_client.post(create_link, data)
    css = theme.css.last()
    assert theme.dirname in str(css.source_file)


def test_created_source_file_name_starts_with_asset_name(
    theme, admin_client, create_link, data
):
    admin_client.post(create_link, data)
    css = theme.css.last()
    source_filename = str(css.source_file).split("/")[-1]
    assert source_filename.startswith("test.")


def test_created_source_file_has_css_extension(theme, admin_client, create_link, data):
    admin_client.post(create_link, data)
    css = theme.css.last()
    source_filename = str(css.source_file).split("/")[-1]
    assert source_filename.endswith(".css")


def test_created_source_file_is_hashed(theme, admin_client, create_link, data):
    admin_client.post(create_link, data)
    css = theme.css.last()
    source_filename = str(css.source_file).split("/")[-1]
    assert ".%s." % css.source_hash in source_filename


def test_css_creation_fails_if_name_is_not_given(
    theme, admin_client, create_link, data
):
    data["name"] = ""
    admin_client.post(create_link, data)
    assert not theme.css.exists()


def test_css_creation_fails_if_name_is_not_valid(
    theme, admin_client, create_link, data
):
    data["name"] = "missing-extension"
    admin_client.post(create_link, data)
    assert not theme.css.exists()


def test_css_creation_fails_if_name_is_already_taken_by_other_css_in_theme(
    theme, admin_client, create_link, data, css
):
    data["name"] = css.name
    admin_client.post(create_link, data)
    assert theme.css.count() == 1


def test_css_name_usage_check_passess_if_name_is_used_by_other_theme_css(
    other_theme, admin_client, data, css
):
    create_link = reverse(
        "misago:admin:themes:new-css-file", kwargs={"pk": other_theme.pk}
    )
    data["name"] = css.name
    admin_client.post(create_link, data)
    assert other_theme.css.exists()


def test_css_creation_fails_if_source_is_not_given(
    theme, admin_client, create_link, data
):
    data["source"] = ""
    admin_client.post(create_link, data)
    assert not theme.css.exists()


def test_css_file_without_url_is_created_without_rebuilding_flag(
    theme, admin_client, create_link, data
):
    admin_client.post(create_link, data)
    css = theme.css.last()
    assert not css.source_needs_building


def test_css_file_with_image_url_is_created_with_rebuilding_flag(
    theme, admin_client, create_link, data, image
):
    data["source"] = "body { background-image: url(/static/%s); }" % image.name
    admin_client.post(create_link, data)
    css = theme.css.last()
    assert css.source_needs_building


def test_creating_css_file_without_image_url_doesnt_trigger_single_css_file_rebuild(
    theme, admin_client, create_link, data, image, mock_build_single_theme_css
):
    admin_client.post(create_link, data)
    mock_build_single_theme_css.assert_not_called()


def test_creating_css_file_with_image_url_triggers_single_css_file_rebuild(
    theme, admin_client, create_link, data, image, mock_build_single_theme_css
):
    data["source"] = "body { background-image: url(/static/%s); }" % image.name
    admin_client.post(create_link, data)
    css = theme.css.last()

    mock_build_single_theme_css.assert_called_once_with(css.pk)


def test_css_file_is_created_with_correct_order(
    theme, admin_client, create_link, css_link, data
):
    admin_client.post(create_link, data)
    css = theme.css.get(name=data["name"])
    assert css.order == 1


def test_creating_css_file_invalidates_theme_cache(admin_client, create_link, data):
    with assert_invalidates_cache(THEME_CACHE):
        admin_client.post(create_link, data)


def test_error_message_is_set_if_user_attempts_to_create_css_in_default_theme(
    default_theme, admin_client
):
    create_link = reverse(
        "misago:admin:themes:new-css-file", kwargs={"pk": default_theme.pk}
    )
    response = admin_client.get(create_link)
    assert_has_error_message(response)


def test_error_message_is_set_if_user_attempts_to_create_css_in_nonexisting_theme(
    nonexisting_theme, admin_client
):
    create_link = reverse(
        "misago:admin:themes:new-css-file", kwargs={"pk": nonexisting_theme.pk}
    )
    response = admin_client.get(create_link)
    assert_has_error_message(response)


def test_css_creation_form_redirects_user_to_edition_after_creation(
    theme, admin_client, create_link, data
):
    data["stay"] = "1"
    response = admin_client.post(create_link, data)
    assert response["location"] == reverse(
        "misago:admin:themes:edit-css-file",
        kwargs={"pk": theme.pk, "css_pk": theme.css.last().pk},
    )


def test_css_edition_form_is_displayed(admin_client, edit_link, css):
    response = admin_client.get(edit_link)
    assert response.status_code == 200
    assert_contains(response, css.name)


def test_css_edition_form_contains_source_file_contents(admin_client, edit_link, css):
    response = admin_client.get(edit_link)
    assert_contains(response, css.source_file.read().decode("utf-8"))


def test_css_name_can_be_changed(admin_client, edit_link, css, data):
    data["name"] = "new-name.css"
    admin_client.post(edit_link, data)

    css.refresh_from_db()
    assert css.name == data["name"]


def test_css_name_change_also_changes_source_file_name(
    admin_client, edit_link, css, data
):
    data["name"] = "new-name.css"
    admin_client.post(edit_link, data)

    css.refresh_from_db()
    assert "new-name" in str(css.source_file)


def test_css_source_can_be_changed(admin_client, edit_link, css, data):
    data["source"] = ".misago-footer { display: none; }"
    admin_client.post(edit_link, data)

    css.refresh_from_db()
    assert css.source_file.read().decode("utf-8") == data["source"]


def test_changing_css_source_also_changes_source_hash(
    admin_client, edit_link, css, data
):
    original_hash = css.source_hash
    data["source"] = ".misago-footer { display: none; }"
    admin_client.post(edit_link, data)

    css.refresh_from_db()
    assert css.source_hash != original_hash


def test_changing_css_source_also_changes_hash_in_filename(
    admin_client, edit_link, css, data
):
    original_hash = css.source_hash
    data["source"] = ".misago-footer { display: none; }"
    admin_client.post(edit_link, data)

    css.refresh_from_db()
    assert original_hash not in str(css.source_file)
    assert css.source_hash in str(css.source_file)


def test_hash_stays_same_if_source_is_not_changed(admin_client, edit_link, css, data):
    original_hash = css.source_hash
    data["name"] = "changed.css"
    data["source"] = css.source_file.read().decode("utf-8")
    admin_client.post(edit_link, data)

    css.refresh_from_db()
    assert original_hash == css.source_hash


def test_file_is_not_updated_if_form_data_has_no_changes(
    admin_client, edit_link, css, data
):
    original_source_file = str(css.source_file)
    data["name"] = css.name
    data["source"] = css.source_file.read().decode("utf-8")
    admin_client.post(edit_link, data)

    css.refresh_from_db()
    assert original_source_file == str(css.source_file)


def test_adding_image_url_to_edited_file_sets_rebuilding_flag(
    theme, admin_client, edit_link, css, data, image
):
    data["source"] = "body { background-image: url(/static/%s); }" % image.name
    admin_client.post(edit_link, data)
    css.refresh_from_db()
    assert css.source_needs_building


def test_removing_url_from_edited_file_removes_rebuilding_flag(
    theme, admin_client, edit_link, css, data
):
    css.source_needs_building = True
    css.save()

    data["source"] = "body { background-image: none; }"
    admin_client.post(edit_link, data)

    css.refresh_from_db()
    assert not css.source_needs_building


def test_adding_image_url_to_edited_file_triggers_single_css_file_rebuild(
    theme, admin_client, edit_link, css, data, image, mock_build_single_theme_css
):
    data["source"] = "body { background-image: url(/static/%s); }" % image.name
    admin_client.post(edit_link, data)

    mock_build_single_theme_css.assert_called_once_with(css.pk)


def test_removing_url_from_edited_file_deosnt_trigger_single_css_file_rebuild(
    theme, admin_client, edit_link, css, data, mock_build_single_theme_css
):
    data["source"] = "body { background-image: none; }"
    admin_client.post(edit_link, data)
    mock_build_single_theme_css.assert_not_called()


def test_css_order_stays_the_same_after_edit(admin_client, edit_link, css, data):
    original_order = css.order
    data["name"] = "changed.css"
    admin_client.post(edit_link, data)

    css.refresh_from_db()
    assert css.order == original_order


def test_editing_css_file_invalidates_theme_cache(admin_client, edit_link, css, data):
    with assert_invalidates_cache(THEME_CACHE):
        admin_client.post(edit_link, data)


def test_css_edit_form_redirects_user_to_edition_after_saving(
    theme, admin_client, edit_link, css, data
):
    data["stay"] = "1"
    response = admin_client.post(edit_link, data)
    assert response["location"] == edit_link


def test_error_message_is_set_if_user_attempts_to_edit_css_file_in_default_theme(
    default_theme, admin_client
):
    edit_link = reverse(
        "misago:admin:themes:edit-css-file",
        kwargs={"pk": default_theme.pk, "css_pk": 1},
    )
    response = admin_client.get(edit_link)
    assert_has_error_message(response)


def test_error_message_is_set_if_user_attempts_to_edit_css_file_in_nonexisting_theme(
    nonexisting_theme, admin_client
):
    edit_link = reverse(
        "misago:admin:themes:edit-css-file",
        kwargs={"pk": nonexisting_theme.pk, "css_pk": 1},
    )
    response = admin_client.get(edit_link)
    assert_has_error_message(response)


def test_error_message_is_set_if_user_attempts_to_edit_css_belonging_to_other_theme(
    other_theme, admin_client, css
):
    edit_link = reverse(
        "misago:admin:themes:edit-css-file",
        kwargs={"pk": other_theme.pk, "css_pk": css.pk},
    )
    response = admin_client.get(edit_link)
    assert_has_error_message(response)


def test_error_message_is_set_if_user_attempts_to_edit_nonexisting_css(
    theme, admin_client
):
    edit_link = reverse(
        "misago:admin:themes:edit-css-file", kwargs={"pk": theme.pk, "css_pk": 1}
    )
    response = admin_client.get(edit_link)
    assert_has_error_message(response)


def test_error_message_is_set_if_user_attempts_to_edit_css_link_with_file_form(
    theme, admin_client, css_link
):
    edit_link = reverse(
        "misago:admin:themes:edit-css-file",
        kwargs={"pk": theme.pk, "css_pk": css_link.pk},
    )
    response = admin_client.get(edit_link)
    assert_has_error_message(response)



from django.db import transaction
from django.utils import timezone
from django.utils.translation import gettext as _

from .exceptions import ModerationError

__all__ = [
    "approve_post",
    "protect_post",
    "unprotect_post",
    "unhide_post",
    "hide_post",
    "delete_post",
]


def approve_post(user, post):
    if not post.is_unapproved:
        return False

    post.is_unapproved = False
    post.save(update_fields=["is_unapproved"])
    return True


def protect_post(user, post):
    if post.is_protected:
        return False

    post.is_protected = True
    post.save(update_fields=["is_protected"])
    if post.is_best_answer:
        post.thread.best_answer_is_protected = True
        post.thread.save(update_fields=["best_answer_is_protected"])
    return True


def unprotect_post(user, post):
    if not post.is_protected:
        return False

    post.is_protected = False
    post.save(update_fields=["is_protected"])
    if post.is_best_answer:
        post.thread.best_answer_is_protected = False
        post.thread.save(update_fields=["best_answer_is_protected"])
    return True


def unhide_post(user, post):
    if post.is_first_post:
        raise ModerationError(
            _("You can't make original post visible without revealing thread.")
        )

    if not post.is_hidden:
        return False

    post.is_hidden = False
    post.save(update_fields=["is_hidden"])
    return True


def hide_post(user, post):
    if post.is_first_post:
        raise ModerationError(_("You can't hide original post without hiding thread."))

    if post.is_hidden:
        return False

    post.is_hidden = True
    post.hidden_by = user
    post.hidden_by_name = user.username
    post.hidden_by_slug = user.slug
    post.hidden_on = timezone.now()
    post.save(
        update_fields=[
            "is_hidden",
            "hidden_by",
            "hidden_by_name",
            "hidden_by_slug",
            "hidden_on",
        ]
    )
    return True


@transaction.atomic
def delete_post(user, post):
    if post.is_first_post:
        raise ModerationError(
            _("You can't delete original post without deleting thread.")
        )

    post.delete()
    return True



from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils import timezone


class AuditTrail(models.Model):
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    created_on = models.DateTimeField(db_index=True, default=timezone.now)
    ip_address = models.GenericIPAddressField()

    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey("content_type", "object_id")

    class Meta:
        ordering = ["-pk"]



from rest_framework.routers import DefaultRouter, DynamicRoute, Route


class MisagoApiRouter(DefaultRouter):
    include_root_view = False
    include_format_suffixes = False

    routes = [
        # List route.
        Route(
            url=r"^{prefix}{trailing_slash}$",
            mapping={"get": "list", "post": "create"},
            name="{basename}-list",
            initkwargs={"suffix": "List"},
            detail=False,
        ),
        # Dynamically generated list routes.
        # Generated using @list_route decorator
        # on methods of the viewset.
        DynamicRoute(
            url=r"^{prefix}/{url_path}{trailing_slash}$",
            name="{basename}-{url_name}",
            initkwargs={},
            detail=False,
        ),
        # Detail route.
        Route(
            url=r"^{prefix}/{lookup}{trailing_slash}$",
            mapping={
                "get": "retrieve",
                "put": "update",
                "patch": "partial_update",
                "delete": "destroy",
            },
            name="{basename}-detail",
            initkwargs={"suffix": "Instance"},
            detail=True,
        ),
        # Dynamically generated detail routes.
        # Generated using @detail_route decorator on methods of the viewset.
        DynamicRoute(
            url=r"^{prefix}/{lookup}/{url_path}{trailing_slash}$",
            name="{basename}-{url_name}",
            initkwargs={},
            detail=True,
        ),
    ]



from django.shortcuts import get_object_or_404
from django.utils.translation import gettext as _

from ...acl.objectacl import add_acl_to_obj
from ...categories import PRIVATE_THREADS_ROOT_NAME, THREADS_ROOT_NAME
from ...categories.models import Category
from ...core.shortcuts import validate_slug
from ...core.viewmodel import ViewModel as BaseViewModel
from ...readtracker.threadstracker import make_read_aware
from ..models import Poll, Thread
from ..participants import make_participants_aware
from ..permissions import (
    allow_see_private_thread,
    allow_see_thread,
    allow_use_private_threads,
)
from ..serializers import PrivateThreadSerializer, ThreadSerializer
from ..subscriptions import make_subscription_aware
from ..threadtypes import trees_map

__all__ = ["ForumThread", "PrivateThread"]

BASE_RELATIONS = [
    "category",
    "poll",
    "starter",
    "starter__rank",
    "starter__ban_cache",
    "starter__online_tracker",
]


class ViewModel(BaseViewModel):
    def __init__(
        self,
        request,
        pk,
        slug=None,
        path_aware=False,
        read_aware=False,
        subscription_aware=False,
        poll_votes_aware=False,
    ):
        model = self.get_thread(request, pk, slug)

        if path_aware:
            model.path = self.get_thread_path(model.category)

        add_acl_to_obj(request.user_acl, model.category)
        add_acl_to_obj(request.user_acl, model)

        if read_aware:
            make_read_aware(request, model)
        if subscription_aware:
            make_subscription_aware(request.user, model)

        self._model = model

        try:
            self._poll = model.poll
            add_acl_to_obj(request.user_acl, self._poll)

            if poll_votes_aware:
                self._poll.make_choices_votes_aware(request.user)
        except Poll.DoesNotExist:
            self._poll = None

    @property
    def poll(self):
        return self._poll

    def get_thread(self, request, pk, slug=None):
        raise NotImplementedError(
            "Thread view model has to implement get_thread(request, pk, slug=None)"
        )

    def get_thread_path(self, category):
        thread_path = []

        if category.level:
            categories = Category.objects.filter(
                tree_id=category.tree_id, lft__lte=category.lft, rght__gte=category.rght
            ).order_by("level")
            thread_path = list(categories)
        else:
            thread_path = [category]

        thread_path[0].name = self.get_root_name()
        return thread_path

    def get_root_name(self):
        raise NotImplementedError("Thread view model has to implement get_root_name()")

    def get_frontend_context(self):
        raise NotImplementedError(
            "Thread view model has to implement get_frontend_context()"
        )

    def get_template_context(self):
        return {
            "thread": self._model,
            "poll": self._poll,
            "category": self._model.category,
            "breadcrumbs": self._model.path,
        }


class ForumThread(ViewModel):
    def get_thread(self, request, pk, slug=None):
        thread = get_object_or_404(
            Thread.objects.select_related(*BASE_RELATIONS),
            pk=pk,
            category__tree_id=trees_map.get_tree_id_for_root(THREADS_ROOT_NAME),
        )

        allow_see_thread(request.user_acl, thread)
        if slug:
            validate_slug(thread, slug)
        return thread

    def get_root_name(self):
        return _("Threads")

    def get_frontend_context(self):
        return ThreadSerializer(self._model).data


class PrivateThread(ViewModel):
    def get_thread(self, request, pk, slug=None):
        allow_use_private_threads(request.user_acl)

        thread = get_object_or_404(
            Thread.objects.select_related(*BASE_RELATIONS),
            pk=pk,
            category__tree_id=trees_map.get_tree_id_for_root(PRIVATE_THREADS_ROOT_NAME),
        )

        make_participants_aware(request.user, thread)
        allow_see_private_thread(request.user_acl, thread)

        if slug:
            validate_slug(thread, slug)

        return thread

    def get_root_name(self):
        return _("Private threads")

    def get_frontend_context(self):
        return PrivateThreadSerializer(self._model).data



import json

from django import template
from django.utils.safestring import mark_safe

from ..utils import encode_json_html

register = template.Library()


@register.filter
def as_json(value):
    json_dump = json.dumps(value)
    # fixes XSS as described in #651
    return mark_safe(encode_json_html(json_dump))



from django.db import migrations

_ = lambda s: s


def create_default_roles(apps, schema_editor):
    Role = apps.get_model("misago_acl", "Role")

    Role.objects.create(
        name=_("Member"),
        special_role="authenticated",
        permissions={
            # account
            "misago.users.permissions.account": {
                "name_changes_allowed": 2,
                "name_changes_expire": 180,
                "can_have_signature": 0,
                "allow_signature_links": 0,
                "allow_signature_images": 0,
            },
            # profiles
            "misago.users.permissions.profiles": {
                "can_browse_users_list": 1,
                "can_search_users": 1,
                "can_follow_users": 1,
                "can_be_blocked": 1,
                "can_see_users_name_history": 0,
                "can_see_users_emails": 0,
                "can_see_users_ips": 0,
                "can_see_hidden_users": 0,
            },
            # attachments
            "misago.threads.permissions.attachments": {
                "max_attachment_size": 4 * 1024,
                "can_download_other_users_attachments": True,
            },
            # polls
            "misago.threads.permissions.polls": {
                "can_start_polls": 1,
                "can_edit_polls": 1,
            },
            # search
            "misago.search.permissions": {"can_search": 1},
        },
    )

    Role.objects.create(
        name=_("Guest"),
        special_role="anonymous",
        permissions={
            # account
            "misago.users.permissions.account": {
                "name_changes_allowed": 0,
                "name_changes_expire": 0,
                "can_have_signature": 0,
                "allow_signature_links": 0,
                "allow_signature_images": 0,
            },
            # profiles
            "misago.users.permissions.profiles": {
                "can_browse_users_list": 1,
                "can_search_users": 1,
                "can_see_users_name_history": 0,
                "can_see_users_emails": 0,
                "can_see_users_ips": 0,
                "can_see_hidden_users": 0,
            },
            # attachments
            "misago.threads.permissions.attachments": {
                "can_download_other_users_attachments": True
            },
            # search
            "misago.search.permissions": {"can_search": 1},
        },
    )

    Role.objects.create(
        name=_("Moderator"),
        permissions={
            # account
            "misago.users.permissions.account": {
                "name_changes_allowed": 5,
                "name_changes_expire": 14,
                "can_have_signature": 1,
                "allow_signature_links": 1,
                "allow_signature_images": 0,
            },
            # profiles
            "misago.users.permissions.profiles": {
                "can_browse_users_list": 1,
                "can_search_users": 1,
                "can_be_blocked": 0,
                "can_see_users_name_history": 1,
                "can_see_ban_details": 1,
                "can_see_users_emails": 1,
                "can_see_users_ips": 1,
                "can_see_hidden_users": 1,
            },
            # attachments
            "misago.threads.permissions.attachments": {
                "max_attachment_size": 8 * 1024,
                "can_download_other_users_attachments": True,
                "can_delete_other_users_attachments": True,
            },
            # polls
            "misago.threads.permissions.polls": {
                "can_start_polls": 2,
                "can_edit_polls": 2,
                "can_delete_polls": 2,
                "can_always_see_poll_voters": 1,
            },
            # moderation
            "misago.threads.permissions.threads": {
                "can_see_unapproved_content_lists": True,
                "can_see_reported_content_lists": True,
                "can_omit_flood_protection": True,
            },
            "misago.users.permissions.moderation": {
                "can_warn_users": 1,
                "can_moderate_avatars": 1,
                "can_moderate_signatures": 1,
                "can_moderate_profile_details": 1,
            },
            # delete users
            "misago.users.permissions.delete": {
                "can_delete_users_newer_than": 0,
                "can_delete_users_with_less_posts_than": 0,
            },
        },
    )

    Role.objects.create(
        name=_("Renaming users"),
        permissions={
            # rename users
            "misago.users.permissions.moderation": {"can_rename_users": 1}
        },
    )

    Role.objects.create(
        name=_("Banning users"),
        permissions={
            # ban users
            "misago.users.permissions.profiles": {"can_see_ban_details": 1},
            "misago.users.permissions.moderation": {
                "can_ban_users": 1,
                "max_ban_length": 14,
                "can_lift_bans": 1,
                "max_lifted_ban_length": 14,
            },
        },
    )

    Role.objects.create(
        name=_("Deleting users"),
        permissions={
            # delete users
            "misago.users.permissions.delete": {
                "can_delete_users_newer_than": 3,
                "can_delete_users_with_less_posts_than": 7,
            }
        },
    )

    Role.objects.create(
        name=_("Can't be blocked"),
        permissions={
            # profiles
            "misago.users.permissions.profiles": {"can_be_blocked": 0}
        },
    )

    Role.objects.create(
        name=_("Private threads"),
        permissions={
            # private threads
            "misago.threads.permissions.privatethreads": {
                "can_use_private_threads": 1,
                "can_start_private_threads": 1,
                "max_private_thread_participants": 3,
                "can_add_everyone_to_private_threads": 0,
                "can_report_private_threads": 1,
                "can_moderate_private_threads": 0,
            }
        },
    )

    Role.objects.create(
        name=_("Private threads moderator"),
        permissions={
            # private threads
            "misago.threads.permissions.privatethreads": {
                "can_use_private_threads": 1,
                "can_start_private_threads": 1,
                "max_private_thread_participants": 15,
                "can_add_everyone_to_private_threads": 1,
                "can_report_private_threads": 1,
                "can_moderate_private_threads": 1,
            }
        },
    )


class Migration(migrations.Migration):

    dependencies = [("misago_acl", "0002_acl_version_tracker")]

    operations = [migrations.RunPython(create_default_roles)]



from django.urls import path
from django.utils.translation import gettext_lazy as _

from .views import (
    AgreementsList,
    DeleteAgreement,
    DisableAgreement,
    EditAgreement,
    NewAgreement,
    SetAgreementAsActive,
)


class MisagoAdminExtension:
    def register_urlpatterns(self, urlpatterns):
        # Legal Agreements
        urlpatterns.namespace("agreements/", "agreements", "settings")
        urlpatterns.patterns(
            "settings:agreements",
            path("", AgreementsList.as_view(), name="index"),
            path("<int:page>/", AgreementsList.as_view(), name="index"),
            path("new/", NewAgreement.as_view(), name="new"),
            path("edit/<int:pk>/", EditAgreement.as_view(), name="edit"),
            path("delete/<int:pk>/", DeleteAgreement.as_view(), name="delete"),
            path(
                "set-as-active/<int:pk>/",
                SetAgreementAsActive.as_view(),
                name="set-as-active",
            ),
            path("disable/<int:pk>/", DisableAgreement.as_view(), name="disable"),
        )

    def register_navigation_nodes(self, site):
        site.add_node(
            name=_("Legal agreements"),
            description=_("Set terms of service and privacy policy contents."),
            parent="settings",
            namespace="agreements",
        )



from datetime import timedelta

from django.urls import reverse
from django.utils import timezone

from ...acl.test import patch_user_acl
from ..models import Poll, PollVote, Thread
from ..test import patch_category_acl
from .test_thread_poll_api import ThreadPollApiTestCase


class ThreadPollDeleteTests(ThreadPollApiTestCase):
    def setUp(self):
        super().setUp()

        self.mock_poll()

    def test_anonymous(self):
        """api requires you to sign in to delete poll"""
        self.logout_user()

        response = self.client.delete(self.api_link)
        self.assertEqual(response.status_code, 403)

    def test_invalid_thread_id(self):
        """api validates that thread id is integer"""
        api_link = reverse(
            "misago:api:thread-poll-detail",
            kwargs={"thread_pk": "kjha6dsa687sa", "pk": self.poll.pk},
        )

        response = self.client.delete(api_link)
        self.assertEqual(response.status_code, 404)

    def test_nonexistant_thread_id(self):
        """api validates that thread exists"""
        api_link = reverse(
            "misago:api:thread-poll-detail",
            kwargs={"thread_pk": self.thread.pk + 1, "pk": self.poll.pk},
        )

        response = self.client.delete(api_link)
        self.assertEqual(response.status_code, 404)

    def test_invalid_poll_id(self):
        """api validates that poll id is integer"""
        api_link = reverse(
            "misago:api:thread-poll-detail",
            kwargs={"thread_pk": self.thread.pk, "pk": "sad98as7d97sa98"},
        )

        response = self.client.delete(api_link)
        self.assertEqual(response.status_code, 404)

    def test_nonexistant_poll_id(self):
        """api validates that poll exists"""
        api_link = reverse(
            "misago:api:thread-poll-detail",
            kwargs={"thread_pk": self.thread.pk, "pk": self.poll.pk + 123},
        )

        response = self.client.delete(api_link)
        self.assertEqual(response.status_code, 404)

    @patch_user_acl({"can_delete_polls": 0})
    def test_no_permission(self):
        """api validates that user has permission to delete poll in thread"""
        response = self.client.delete(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.json(), {"detail": "You can't delete polls."})

    @patch_user_acl({"can_delete_polls": 1, "poll_edit_time": 5})
    def test_no_permission_timeout(self):
        """api validates that user's window to delete poll in thread has closed"""
        self.poll.posted_on = timezone.now() - timedelta(minutes=15)
        self.poll.save()

        response = self.client.delete(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "You can't delete polls that are older than 5 minutes."},
        )

    @patch_user_acl({"can_delete_polls": 1})
    def test_no_permission_poll_closed(self):
        """api validates that user's window to delete poll in thread has closed"""
        self.poll.posted_on = timezone.now() - timedelta(days=15)
        self.poll.length = 5
        self.poll.save()

        response = self.client.delete(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "This poll is over. You can't delete it."}
        )

    @patch_user_acl({"can_delete_polls": 1})
    def test_no_permission_other_user_poll(self):
        """api validates that user has permission to delete other user poll in thread"""
        self.poll.poster = None
        self.poll.save()

        response = self.client.delete(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "You can't delete other users polls in this category."},
        )

    @patch_user_acl({"can_delete_polls": 1})
    @patch_category_acl({"can_close_threads": False})
    def test_no_permission_closed_thread(self):
        """api validates that user has permission to delete poll in closed thread"""
        self.thread.is_closed = True
        self.thread.save()

        response = self.client.delete(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "This thread is closed. You can't delete polls in it."},
        )

    @patch_user_acl({"can_delete_polls": 1})
    @patch_category_acl({"can_close_threads": True})
    def test_closed_thread(self):
        """api validates that user has permission to delete poll in closed thread"""
        self.thread.is_closed = True
        self.thread.save()

        response = self.client.delete(self.api_link)
        self.assertEqual(response.status_code, 200)

    @patch_user_acl({"can_delete_polls": 1})
    @patch_category_acl({"can_close_threads": False})
    def test_no_permission_closed_category(self):
        """api validates that user has permission to delete poll in closed category"""
        self.category.is_closed = True
        self.category.save()

        response = self.client.delete(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "This category is closed. You can't delete polls in it."},
        )

    @patch_user_acl({"can_delete_polls": 1})
    @patch_category_acl({"can_close_threads": True})
    def test_closed_category(self):
        """api validates that user has permission to delete poll in closed category"""
        self.category.is_closed = True
        self.category.save()

        response = self.client.delete(self.api_link)
        self.assertEqual(response.status_code, 200)

    @patch_user_acl({"can_delete_polls": 1, "poll_edit_time": 5})
    def test_poll_delete(self):
        """api deletes poll and associated votes"""
        response = self.client.delete(self.api_link)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), {"can_start_poll": True})

        self.assertEqual(Poll.objects.count(), 0)
        self.assertEqual(PollVote.objects.count(), 0)

        # api set poll flag on thread to False
        thread = Thread.objects.get(pk=self.thread.pk)
        self.assertFalse(thread.has_poll)

    @patch_user_acl({"can_delete_polls": 2, "poll_edit_time": 5})
    def test_other_user_poll_delete(self):
        """api deletes other user's poll and associated votes, even if its over"""
        self.poll.poster = None
        self.poll.posted_on = timezone.now() - timedelta(days=15)
        self.poll.length = 5
        self.poll.save()

        response = self.client.delete(self.api_link)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), {"can_start_poll": True})

        self.assertEqual(Poll.objects.count(), 0)
        self.assertEqual(PollVote.objects.count(), 0)

        # api set poll flag on thread to False
        thread = Thread.objects.get(pk=self.thread.pk)
        self.assertFalse(thread.has_poll)



from django.contrib.auth import get_user_model
from django.urls import reverse
from rest_framework import serializers

from . import RankSerializer
from ...core.serializers import MutableFields

User = get_user_model()

__all__ = ["StatusSerializer", "UserSerializer", "UserCardSerializer"]


class StatusSerializer(serializers.Serializer):
    is_offline = serializers.BooleanField()
    is_online = serializers.BooleanField()
    is_hidden = serializers.BooleanField()
    is_offline_hidden = serializers.BooleanField()
    is_online_hidden = serializers.BooleanField()
    last_click = serializers.DateTimeField()

    is_banned = serializers.BooleanField()
    banned_until = serializers.DateTimeField()


class UserSerializer(serializers.ModelSerializer, MutableFields):
    email = serializers.SerializerMethodField()
    rank = RankSerializer(many=False, read_only=True)
    signature = serializers.SerializerMethodField()

    acl = serializers.SerializerMethodField()
    is_followed = serializers.SerializerMethodField()
    is_blocked = serializers.SerializerMethodField()
    meta = serializers.SerializerMethodField()
    real_name = serializers.SerializerMethodField()
    status = serializers.SerializerMethodField()

    api = serializers.SerializerMethodField()
    url = serializers.SerializerMethodField()

    class Meta:
        model = User
        fields = [
            "id",
            "username",
            "slug",
            "email",
            "joined_on",
            "rank",
            "title",
            "avatars",
            "is_avatar_locked",
            "signature",
            "is_signature_locked",
            "followers",
            "following",
            "threads",
            "posts",
            "acl",
            "is_followed",
            "is_blocked",
            "meta",
            "real_name",
            "status",
            "api",
            "url",
        ]

    def get_acl(self, obj):
        return obj.acl

    def get_email(self, obj):
        request = self.context["request"]
        if obj == request.user or request.user_acl["can_see_users_emails"]:
            return obj.email

    def get_is_followed(self, obj):
        request = self.context["request"]
        if obj.acl["can_follow"]:
            return request.user.is_following(obj)
        return False

    def get_is_blocked(self, obj):
        request = self.context["request"]
        if obj.acl["can_block"]:
            return request.user.is_blocking(obj)
        return False

    def get_meta(self, obj):
        return {"score": obj.score}

    def get_real_name(self, obj):
        return obj.get_real_name()

    def get_signature(self, obj):
        if obj.has_valid_signature:
            return obj.signature_parsed

    def get_status(self, obj):
        try:
            return StatusSerializer(obj.status).data
        except AttributeError:
            return None

    def get_api(self, obj):
        return {
            "index": reverse("misago:api:user-detail", kwargs={"pk": obj.pk}),
            "follow": reverse("misago:api:user-follow", kwargs={"pk": obj.pk}),
            "ban": reverse("misago:api:user-ban", kwargs={"pk": obj.pk}),
            "details": reverse("misago:api:user-details", kwargs={"pk": obj.pk}),
            "edit_details": reverse(
                "misago:api:user-edit-details", kwargs={"pk": obj.pk}
            ),
            "moderate_avatar": reverse(
                "misago:api:user-moderate-avatar", kwargs={"pk": obj.pk}
            ),
            "moderate_username": reverse(
                "misago:api:user-moderate-username", kwargs={"pk": obj.pk}
            ),
            "delete": reverse("misago:api:user-delete", kwargs={"pk": obj.pk}),
            "followers": reverse("misago:api:user-followers", kwargs={"pk": obj.pk}),
            "follows": reverse("misago:api:user-follows", kwargs={"pk": obj.pk}),
            "threads": reverse("misago:api:user-threads", kwargs={"pk": obj.pk}),
            "posts": reverse("misago:api:user-posts", kwargs={"pk": obj.pk}),
        }

    def get_url(self, obj):
        return obj.get_absolute_url()


UserCardSerializer = UserSerializer.subset_fields(
    "id",
    "username",
    "joined_on",
    "rank",
    "title",
    "avatars",
    "followers",
    "threads",
    "posts",
    "real_name",
    "status",
    "url",
)



from pathlib import Path

from django.db import models

from ..core.utils import get_file_hash
from .hydrators import dehydrate_value, hydrate_value


class SettingsManager(models.Manager):
    def change_setting(self, setting, dry_value=None, wet_value=None):
        if dry_value:
            return self.filter(setting=setting).update(dry_value=dry_value)

        if wet_value:
            try:
                setting = self.get(setting=setting)
                setting.value = wet_value
                setting.save(update_fields=["dry_value"])
            except Setting.DoesNotExist:
                return 0


class Setting(models.Model):
    setting = models.CharField(max_length=255, unique=True)
    python_type = models.CharField(max_length=255, default="string")
    dry_value = models.TextField(null=True, blank=True)
    image = models.ImageField(
        upload_to="conf",
        height_field="image_height",
        width_field="image_width",
        null=True,
        blank=True,
    )
    image_size = models.PositiveIntegerField(null=True, blank=True)
    image_width = models.PositiveIntegerField(null=True, blank=True)
    image_height = models.PositiveIntegerField(null=True, blank=True)
    is_public = models.BooleanField(default=False)
    is_lazy = models.BooleanField(default=False)

    objects = SettingsManager()

    @property
    def image_dimensions(self):
        if self.image_width and self.image_height:
            return self.image_width, self.image_height
        return None

    @property
    def value(self):
        if self.python_type == "image":
            return self.image
        return hydrate_value(self.python_type, self.dry_value)

    @value.setter
    def value(self, new_value):
        if new_value is not None:
            if self.python_type == "image":
                rename_image_file(new_value, self.setting)
                self.image = new_value
                self.image_size = new_value.size
            else:
                self.dry_value = dehydrate_value(self.python_type, new_value)
        else:
            self.dry_value = None
        return new_value


def rename_image_file(file_obj, prefix):
    name_parts = [
        prefix.replace("_", "-"),
        get_file_hash(file_obj),
        Path(file_obj.name).suffix.strip(".").lower(),
    ]
    file_obj.name = ".".join(name_parts)



from ....acl.objectacl import add_acl_to_obj
from ....core.apipatch import ApiPatch
from ...moderation import posts as moderation
from ...permissions import allow_hide_event, allow_unhide_event

event_patch_dispatcher = ApiPatch()


def patch_acl(request, event, value):
    """useful little op that updates event acl to current state"""
    if value:
        add_acl_to_obj(request.user_acl, event)
        return {"acl": event.acl}
    return {"acl": None}


event_patch_dispatcher.add("acl", patch_acl)


def patch_is_hidden(request, event, value):
    if value:
        allow_hide_event(request.user_acl, event)
        moderation.hide_post(request.user, event)
    else:
        allow_unhide_event(request.user_acl, event)
        moderation.unhide_post(request.user, event)

    return {"is_hidden": event.is_hidden}


event_patch_dispatcher.replace("is-hidden", patch_is_hidden)


def event_patch_endpoint(request, event):
    old_is_hidden = event.is_hidden

    response = event_patch_dispatcher.dispatch(request, event)

    if old_is_hidden != event.is_hidden:
        event.thread.synchronize()
        event.thread.save()

        event.category.synchronize()
        event.category.save()

    return response



from django import forms
from django.db.models import Q
from django.utils import html

from ...core.utils import format_plaintext_for_html

__all__ = [
    "ProfileField",
    "TextProfileField",
    "UrlProfileField",
    "TextareaProfileField",
    "ChoiceProfileField",
]


class ProfileField:
    """
    Basic profile field
    """

    fieldname = None
    label = None
    help_text = None
    readonly = False

    def is_editable(self, request, user):
        return not self.readonly

    def get_label(self, user):
        if not self.label:
            raise NotImplementedError(
                "profile field class has to define label "
                "attribute or get_label(user) method"
            )
        return self.label

    def get_help_text(self, user):
        return self.help_text

    def get_form_field(self, request, user):
        return None

    def get_form_field_json(self, request, user):
        return {
            "fieldname": self.fieldname,
            "label": self.get_label(user),
            "help_text": self.get_help_text(user),
            "initial": user.profile_fields.get(self.fieldname, ""),
            "input": self.get_input_json(request, user),
        }

    def get_input_json(self, request, user):
        return None

    def clean(self, request, user, data):
        return data

    def get_display_data(self, request, user):
        value = user.profile_fields.get(self.fieldname, "").strip()
        if not self.readonly and not value:
            return None

        data = self.get_value_display_data(request, user, value)
        if not data:
            return None

        data.update({"fieldname": self.fieldname, "name": str(self.get_label(user))})

        return data

    def get_value_display_data(self, request, user, value):
        return {"text": value}

    def search_users(self, criteria):
        if self.readonly:
            return None

        return Q(**{"profile_fields__%s__contains" % self.fieldname: criteria})


class ChoiceProfileField(ProfileField):
    choices = None

    def get_choices(self, user=None):
        if not self.choices:
            raise NotImplementedError(
                "profile field ChoiceProfileField has to define "
                "choices attribute or get_choices(user) method"
            )
        return self.choices

    def get_form_field(self, request, user):
        return forms.ChoiceField(
            label=self.get_label(user),
            help_text=self.get_help_text(user),
            initial=user.profile_fields.get(self.fieldname),
            choices=self.get_choices(user),
            disabled=self.readonly,
            required=False,
        )

    def get_input_json(self, request, user):
        choices = []
        for key, choice in self.get_choices():  # pylint: disable=not-an-iterable
            choices.append({"value": key, "label": choice})

        return {"type": "select", "choices": choices}

    def get_value_display_data(self, request, user, value):
        for key, name in self.get_choices():  # pylint: disable=not-an-iterable
            if key == value:
                return {"text": str(name)}

    def search_users(self, criteria):
        """custom search implementation for choice fields"""
        q_obj = Q(**{"profile_fields__%s__contains" % self.fieldname: criteria})

        for key, choice in self.get_choices():  # pylint: disable=not-an-iterable
            if key and criteria.lower() in str(choice).lower():
                q_obj = q_obj | Q(**{"profile_fields__%s" % self.fieldname: key})

        return q_obj


class TextProfileField(ProfileField):
    def get_form_field(self, request, user):
        return forms.CharField(
            label=self.get_label(user),
            help_text=self.get_help_text(user),
            initial=user.profile_fields.get(self.fieldname),
            max_length=250,
            disabled=self.readonly,
            required=False,
        )

    def get_input_json(self, request, user):
        return {"type": "text"}


class TextareaProfileField(ProfileField):
    def get_form_field(self, request, user):
        return forms.CharField(
            label=self.get_label(user),
            help_text=self.get_help_text(user),
            initial=user.profile_fields.get(self.fieldname),
            max_length=500,
            widget=forms.Textarea(attrs={"rows": 4}),
            disabled=self.readonly,
            required=False,
        )

    def get_input_json(self, request, user):
        return {"type": "textarea"}

    def get_value_display_data(self, request, user, value):
        return {"html": html.linebreaks(html.escape(value))}


class UrlifiedTextareaProfileField(TextareaProfileField):
    def get_value_display_data(self, request, user, value):
        return {"html": format_plaintext_for_html(value)}


class UrlProfileField(TextProfileField):
    def get_form_field(self, request, user):
        return forms.URLField(
            label=self.get_label(user),
            help_text=self.get_help_text(user),
            initial=user.profile_fields.get(self.fieldname),
            max_length=250,
            disabled=self.readonly,
            required=False,
        )

    def get_value_display_data(self, request, user, value):
        return {"text": value, "url": value}



from pathlib import Path

from django.core.exceptions import ValidationError
from django.utils.translation import gettext as _
from PIL import Image

from . import store
from ...conf import settings

ALLOWED_EXTENSIONS = (".gif", ".png", ".jpg", ".jpeg")
ALLOWED_MIME_TYPES = ("image/gif", "image/jpeg", "image/png", "image/mpo")


def handle_uploaded_file(request, user, uploaded_file):
    image = validate_uploaded_file(request.settings, uploaded_file)
    store.store_temporary_avatar(user, image)


def validate_uploaded_file(settings, uploaded_file):
    try:
        validate_file_size(settings, uploaded_file)
        validate_extension(uploaded_file)
        validate_mime(uploaded_file)
        return validate_dimensions(uploaded_file)
    except ValidationError as e:
        try:
            temporary_file_path = Path(uploaded_file.temporary_file_path())
            if temporary_file_path.exists():
                temporary_file_path.unlink()
        except Exception:  # pylint: disable=broad-except
            pass
        raise e


def validate_file_size(settings, uploaded_file):
    upload_limit = settings.avatar_upload_limit * 1024
    if uploaded_file.size > upload_limit:
        raise ValidationError(_("Uploaded file is too big."))


def validate_extension(uploaded_file):
    lowercased_name = uploaded_file.name.lower()
    for extension in ALLOWED_EXTENSIONS:
        if lowercased_name.endswith(extension):
            return True
    raise ValidationError(_("Uploaded file type is not allowed."))


def validate_mime(uploaded_file):
    if uploaded_file.content_type not in ALLOWED_MIME_TYPES:
        raise ValidationError(_("Uploaded file type is not allowed."))


def validate_dimensions(uploaded_file):
    image = Image.open(uploaded_file)

    min_size = max(settings.MISAGO_AVATARS_SIZES)
    if min(image.size) < min_size:
        message = _("Uploaded image should be at least %(size)s pixels tall and wide.")
        raise ValidationError(message % {"size": min_size})

    if image.size[0] * image.size[1] > 2000 * 3000:
        message = _("Uploaded image is too big.")
        raise ValidationError(message)

    image_ratio = float(min(image.size)) / float(max(image.size))
    if image_ratio < 0.25:
        message = _("Uploaded image ratio cannot be greater than 16:9.")
        raise ValidationError(message)
    return image


def clean_crop(image, crop):
    message = _("Crop data is invalid. Please try again.")

    crop_dict = {}
    try:
        crop_dict = {
            "x": float(crop["offset"]["x"]),
            "y": float(crop["offset"]["y"]),
            "zoom": float(crop["zoom"]),
        }
    except (KeyError, TypeError, ValueError):
        raise ValidationError(message)

    if crop_dict["zoom"] < 0 or crop_dict["zoom"] > 1:
        raise ValidationError(message)

    min_size = max(settings.MISAGO_AVATARS_SIZES)

    zoomed_size = (
        round(float(image.size[0]) * crop_dict["zoom"], 2),
        round(float(image.size[1]) * crop_dict["zoom"], 2),
    )

    if min(zoomed_size) < min_size:
        raise ValidationError(message)

    crop_square = {"x": crop_dict["x"] * -1, "y": crop_dict["y"] * -1}

    if crop_square["x"] < 0 or crop_square["y"] < 0:
        raise ValidationError(message)

    if crop_square["x"] + min_size > zoomed_size[0]:
        raise ValidationError(message)

    if crop_square["y"] + min_size > zoomed_size[1]:
        raise ValidationError(message)

    return crop_dict


def crop_source_image(user, source, crop):
    if source == "tmp":
        image = Image.open(user.avatar_tmp)
    else:
        image = Image.open(user.avatar_src)
    crop = clean_crop(image, crop)

    min_size = max(settings.MISAGO_AVATARS_SIZES)
    if image.size[0] == min_size and image.size[0] == image.size[1]:
        cropped_image = image
    else:
        upscale = 1.0 / crop["zoom"]
        cropped_image = image.crop(
            (
                int(round(crop["x"] * upscale * -1, 0)),
                int(round(crop["y"] * upscale * -1, 0)),
                int(round((crop["x"] - min_size) * upscale * -1, 0)),
                int(round((crop["y"] - min_size) * upscale * -1, 0)),
            )
        )

    if source == "tmp":
        store.store_new_avatar(user, cropped_image, delete_tmp=False)
        store.store_original_avatar(user)
    else:
        store.store_new_avatar(user, cropped_image, delete_src=False)

    return crop


def has_temporary_avatar(user):
    return bool(user.avatar_tmp)


def has_source_avatar(user):
    return bool(user.avatar_src)



import random
from datetime import timedelta

from django.utils import timezone

from ..users.models import Ban


def get_fake_username_ban(fake):
    ban = _create_base_ban(fake, Ban.USERNAME)

    banned_value = fake.first_name()
    if random.randint(0, 100) < 31:
        banned_value = "%s*" % banned_value
    elif random.randint(0, 100) < 31:
        banned_value = "*%s" % banned_value
    elif random.randint(0, 100) < 31:
        banned_value = list(banned_value)
        banned_value.insert(random.randint(0, len(banned_value) - 1), "*")
        banned_value = "".join(banned_value)

    ban.banned_value = banned_value
    ban.save()
    return ban


def get_fake_email_ban(fake):
    ban = _create_base_ban(fake, Ban.EMAIL)

    if random.randint(0, 100) < 35:
        ban.banned_value = "*@%s" % fake.domain_name()
    else:
        ban.banned_value = fake.email()

    ban.save()
    return ban


def get_fake_ip_ban(fake):
    ban = _create_base_ban(fake, Ban.IP)

    if random.randint(0, 1):
        banned_value = fake.ipv4()
        if random.randint(0, 100) < 35:
            banned_value = banned_value.split(".")
            banned_value = ".".join(banned_value[: random.randint(1, 3)])
            banned_value = "%s.*" % banned_value
        elif random.randint(0, 100) < 35:
            banned_value = banned_value.split(".")
            banned_value = ".".join(banned_value[random.randint(1, 3) :])
            banned_value = "*.%s" % banned_value
        elif random.randint(0, 100) < 35:
            banned_value = banned_value.split(".")
            banned_value[random.randint(0, 3)] = "*"
            banned_value = ".".join(banned_value)
    else:
        banned_value = fake.ipv6()

        if random.randint(0, 100) < 35:
            banned_value = banned_value.split(":")
            banned_value = ":".join(banned_value[: random.randint(1, 7)])
            banned_value = "%s:*" % banned_value
        elif random.randint(0, 100) < 35:
            banned_value = banned_value.split(":")
            banned_value = ":".join(banned_value[: random.randint(1, 7)])
            banned_value = "*:%s" % banned_value
        elif random.randint(0, 100) < 35:
            banned_value = banned_value.split(":")
            banned_value[random.randint(0, 7)] = "*"
            banned_value = ":".join(banned_value)

    ban.banned_value = banned_value
    ban.save()
    return ban


def _create_base_ban(fake, ban_type):
    ban = Ban(check_type=ban_type)

    if random.randint(0, 10) == 0:
        ban.user_message = fake.sentence()

    if random.randint(0, 10) == 0:
        ban.staff_message = fake.sentence()

    if random.randint(0, 1):
        # Lets make ban temporary
        ban_length = timedelta(days=random.randint(0, 300))
        if random.randint(0, 1):
            ban.valid_until = timezone.now().date() - ban_length
        else:
            ban.valid_until = timezone.now().date() + ban_length

    return ban



from django.db import migrations

_ = lambda s: s


def create_default_categories_roles(apps, schema_editor):
    CategoryRole = apps.get_model("misago_categories", "CategoryRole")

    CategoryRole.objects.create(
        name=_("In moderation queue"),
        permissions={
            # threads perms
            "misago.threads.permissions.threads": {
                "require_threads_approval": 1,
                "require_replies_approval": 1,
            }
        },
    )


class Migration(migrations.Migration):

    dependencies = [("misago_categories", "0005_auto_20170303_2027")]

    operations = [migrations.RunPython(create_default_categories_roles)]



import json
import re

from django.core.files.base import ContentFile

from ...core.utils import get_file_hash


def create_css(theme, css):
    order = None
    if css_exists(theme, css):
        order = get_css_order(theme, css)
        delete_css(theme, css)
    save_css(theme, css, order)


def css_exists(theme, css):
    return theme.css.filter(name=css.name).exists()


def get_css_order(theme, css):
    return theme.css.get(name=css.name).order


def delete_css(theme, css):
    theme.css.get(name=css.name).delete()


def save_css(theme, css, order=None):
    if order is None:
        order = get_next_css_order(theme)

    theme.css.create(
        name=css.name,
        source_file=css,
        source_hash=get_file_hash(css),
        source_needs_building=css_needs_rebuilding(css),
        size=css.size,
        order=order,
    )


def css_needs_rebuilding(css):
    css.seek(0)
    css_source = css.read().decode("utf-8")
    return "url(" in css_source


def get_theme_media_map(theme):
    media_map = {}
    for media in theme.media.all():
        escaped_url = json.dumps(media.file.url)
        media_map[media.name] = escaped_url

        media_filename = str(media.file).split("/")[-1]
        media_map[media_filename] = escaped_url
    return media_map


def rebuild_css(media_map, css):
    if css.build_file:
        css.build_file.delete(save=False)

    css_source = css.source_file.read().decode("utf-8")
    build_source = change_css_source(media_map, css_source).encode("utf-8")

    build_file_name = css.name
    if css.source_hash in build_file_name:
        build_file_name = build_file_name.replace(".%s" % css.source_hash, "")
    build_file = ContentFile(build_source, build_file_name)

    css.build_file = build_file
    css.build_hash = get_file_hash(build_file)
    css.size = len(build_source)
    css.save()


CSS_URL_REGEX = re.compile(r"url\((.+)\)")


def change_css_source(media_map, css_source):
    url_replacer = get_url_replacer(media_map)
    return CSS_URL_REGEX.sub(url_replacer, css_source).strip()


def get_url_replacer(media_map):
    def replacer(matchobj):
        url = matchobj.group(1).strip("\"'").strip()
        if is_url_absolute(url):
            return matchobj.group(0)

        media_name = url.split("/")[-1]
        if media_name in media_map:
            return "url(%s)" % media_map[media_name]

        return matchobj.group(0)

    return replacer


def is_url_absolute(url):
    if url.startswith("//") or url.startswith("://"):
        return True

    if url.lower().startswith("https://"):
        return True

    if url.lower().startswith("http://"):
        return True

    return False


def get_next_css_order(theme):
    last_css = theme.css.order_by("order").last()
    if last_css:
        return last_css.order + 1
    return 0


def move_css_up(theme, css):
    previous_css = theme.css.filter(order__lt=css.order).order_by("-order").first()
    if not previous_css:
        return False

    css.order, previous_css.order = previous_css.order, css.order
    css.save(update_fields=["order"])
    previous_css.save(update_fields=["order"])

    return True


def move_css_down(theme, css):
    next_css = theme.css.filter(order__gt=css.order).order_by("order").first()
    if not next_css:
        return False

    css.order, next_css.order = next_css.order, css.order
    css.save(update_fields=["order"])
    next_css.save(update_fields=["order"])

    return True



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots[
    "test_hr_bbcode_is_replaced_if_its_alone_in_paragraph 1"
] = """<p>Lorem ipsum dolor met.</p>
<hr />
<p>Sit amet elit.</p>"""



from django.utils.translation import get_language

from .. import __version__
from ..hooks import context_processors as context_processors_hooks
from .momentjs import get_locale_url


def misago_version(request):
    return {"MISAGO_VERSION": __version__}


def request_path(request):
    return {"request_path": request.path}


def current_link(request):
    if not request.resolver_match or request.frontend_context.get("CURRENT_LINK"):
        return {}

    url_name = request.resolver_match.url_name
    if request.resolver_match.namespaces:
        namespaces = ":".join(request.resolver_match.namespaces)
        link_name = "%s:%s" % (namespaces, url_name)
    else:
        link_name = url_name

    request.frontend_context.update({"CURRENT_LINK": link_name})

    return {}


def momentjs_locale(request):
    return {"MOMENTJS_LOCALE_URL": get_locale_url(get_language())}


def frontend_context(request):
    if request.include_frontend_context:
        return {"frontend_context": request.frontend_context}
    return {}


def hooks(request):
    context = {}
    for hook in context_processors_hooks:
        context.update(hook(request))
    return context



"""
Utility module useful for refactors,

Defines utility function for easy raising of warnings in deprecated functions.

This function is set to use stacklevel 3 as default making following code in test.py:

def depreacted_function():
    warn("This function is deprecated")


def other_function():
    deprecated_function()


Will raise warning about 1st line in other_function calling deprecated function.
"""
import warnings


class RemovedInMisagoWarning(Warning):
    pass


def warn(message, category=RemovedInMisagoWarning, stacklevel=0):
    warnings.warn(message, category, stacklevel + 3)



# Generated by Django 2.2.1 on 2019-05-18 16:59

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("misago_users", "0019_remove_custom_partial_indexes")]

    operations = [
        migrations.AddIndex(
            model_name="user",
            index=models.Index(
                condition=models.Q(is_staff=True),
                fields=["is_staff"],
                name="misago_user_is_staff_part",
            ),
        ),
        migrations.AddIndex(
            model_name="user",
            index=models.Index(
                condition=models.Q(requires_activation__gt=0),
                fields=["requires_activation"],
                name="misago_user_requires_acti_part",
            ),
        ),
        migrations.AddIndex(
            model_name="user",
            index=models.Index(
                condition=models.Q(is_deleting_account=True),
                fields=["is_deleting_account"],
                name="misago_user_is_deleting_a_part",
            ),
        ),
    ]



from django.db import migrations

from ...core.utils import slugify

_ = lambda s: s


def create_default_categories_tree(apps, schema_editor):
    Category = apps.get_model("misago_categories", "Category")

    Category.objects.create(
        special_role="private_threads",
        name="Private",
        slug="private",
        lft=1,
        rght=2,
        tree_id=0,
        level=0,
    )

    root = Category.objects.create(
        special_role="root_category",
        name="Root",
        slug="root",
        lft=3,
        rght=6,
        tree_id=1,
        level=0,
    )

    category_name = _("First category")

    Category.objects.create(
        parent=root,
        lft=4,
        rght=5,
        tree_id=1,
        level=1,
        name=category_name,
        slug=slugify(category_name),
    )


class Migration(migrations.Migration):

    dependencies = [("misago_categories", "0001_initial")]

    operations = [migrations.RunPython(create_default_categories_tree)]



from django.urls import reverse

from ....cache.test import assert_invalidates_cache
from ....test import assert_contains
from ... import SOCIALAUTH_CACHE
from ...models import SocialAuthProvider


def test_form_displays_for_enabled_provider(admin_client, provider):
    response = admin_client.get(
        reverse("misago:admin:settings:socialauth:edit", kwargs={"pk": provider.pk})
    )
    assert response.status_code == 200


def test_form_displays_for_disabled_provider(admin_client, disabled_provider):
    response = admin_client.get(
        reverse(
            "misago:admin:settings:socialauth:edit", kwargs={"pk": disabled_provider.pk}
        )
    )
    assert response.status_code == 200


def test_form_displays_for_unset_provider(admin_client):
    response = admin_client.get(
        reverse("misago:admin:settings:socialauth:edit", kwargs={"pk": "facebook"})
    )
    assert response.status_code == 200


def test_form_handles_undefined_provider(admin_client):
    response = admin_client.get(
        reverse("misago:admin:settings:socialauth:edit", kwargs={"pk": "undefined"})
    )
    assert response.status_code == 302


def test_form_displays_provider_settings(admin_client, provider):
    provider.settings = {"key": "test-key", "secret": "test-secret"}
    provider.save()

    response = admin_client.get(
        reverse("misago:admin:settings:socialauth:edit", kwargs={"pk": provider.pk})
    )

    assert_contains(response, "test-key")
    assert_contains(response, "test-secret")


def test_form_updates_provider_settings(admin_client, provider):
    admin_client.post(
        reverse("misago:admin:settings:socialauth:edit", kwargs={"pk": provider.pk}),
        {"key": "test-key", "secret": "test-secret"},
    )

    provider.refresh_from_db()
    assert provider.settings == {
        "associate_by_email": 0,
        "key": "test-key",
        "secret": "test-secret",
    }


def test_form_updates_provider_button_text(admin_client, provider):
    admin_client.post(
        reverse("misago:admin:settings:socialauth:edit", kwargs={"pk": provider.pk}),
        {"button_text": "Hello world!"},
    )

    provider.refresh_from_db()
    assert provider.button_text == "Hello world!"


def test_form_updates_provider_button_color(admin_client, provider):
    admin_client.post(
        reverse("misago:admin:settings:socialauth:edit", kwargs={"pk": provider.pk}),
        {"button_color": "#ff00ff"},
    )

    provider.refresh_from_db()
    assert provider.button_color == "#ff00ff"


def test_form_invalidates_providers_cache(admin_client, provider):
    with assert_invalidates_cache(SOCIALAUTH_CACHE):
        admin_client.post(
            reverse(
                "misago:admin:settings:socialauth:edit", kwargs={"pk": provider.pk}
            ),
            {"is_active": "0"},
        )


def test_form_sets_provider_order_on_setup(admin_client, other_provider):
    admin_client.post(
        reverse("misago:admin:settings:socialauth:edit", kwargs={"pk": "facebook"}),
        {"is_active": "1", "key": "test-key", "secret": "test-secret"},
    )

    provider = SocialAuthProvider.objects.get(pk="facebook")
    assert provider.order == other_provider.order + 1



from django import forms, template
from django.utils.html import format_html_join

from ..forms import YesNoSwitchBase

register = template.Library()


@register.inclusion_tag("misago/admin/form/row.html")
def form_row(field, label_class=None, field_class=None):
    return {"field": field, "label_class": label_class, "field_class": field_class}


@register.inclusion_tag("misago/admin/form/image_row.html")
def form_image_row(
    field,
    label_class=None,
    field_class=None,
    image_class=None,
    delete_field=None,
    size=None,
    dimensions=None,
):
    return {
        "field": field,
        "field_image": field.initial,
        "size": size,
        "dimensions": get_field_image_dimensions(dimensions),
        "delete_field": delete_field,
        "label_class": label_class,
        "field_class": field_class,
        "image_class": image_class,
    }


def get_field_image_dimensions(dimensions):
    if dimensions:
        return {"width": dimensions[0], "height": dimensions[1]}
    return None


@register.inclusion_tag("misago/admin/form/checkbox_row.html")
def form_checkbox_row(field, label_class=None, field_class=None):
    return {"field": field, "label_class": label_class, "field_class": field_class}


@register.inclusion_tag("misago/admin/form/input.html")
def form_input(field):
    attrs = field.field.widget.attrs
    context = field.field.widget.get_context(field.html_name, field.value(), attrs)
    context["attrs"] = attrs
    context["field"] = field
    return context


@register.simple_tag
def render_attrs(widget, class_name=None):
    rendered_attrs = []
    for attr, value in widget["attrs"].items():
        if value not in (True, False, None):
            rendered_attrs.append((attr, value))
    if not widget["attrs"].get("class") and class_name:
        rendered_attrs.append(("class", class_name))
    return format_html_join(" ", '{}="{}"', rendered_attrs)


@register.simple_tag
def render_bool_attrs(widget):
    attrs_html = []
    for attr, value in widget.items():
        if value is True:
            attrs_html.append(attr)
    return " ".join(attrs_html)


@register.filter
def is_yesno_switch_field(field):
    return isinstance(field.field, YesNoSwitchBase)


@register.filter
def is_radio_select_field(field):
    return isinstance(field.field.widget, forms.RadioSelect)


@register.filter
def is_select_field(field):
    return isinstance(field.field.widget, forms.Select)


@register.filter
def is_multiple_choice_field(field):
    multichoice_widgets = (forms.CheckboxSelectMultiple, forms.SelectMultiple)
    return isinstance(field.field.widget, multichoice_widgets)


@register.filter
def is_textarea_field(field):
    return isinstance(field.field.widget, forms.Textarea)


@register.filter
def get_options(field):
    """Filter that extracts field choices into an easily iterable list"""
    widget = field.field.widget
    attrs = dict(id=field.auto_id, **widget.attrs)
    context = widget.get_context(field.html_name, field.value(), attrs)
    widget_context = context["widget"]
    return widget.options(widget_context["name"], widget_context["value"], attrs)



from django.contrib.auth import get_user_model
from django.core.exceptions import PermissionDenied
from django.utils.translation import gettext as _
from django.utils.translation import gettext_lazy

from ..search import SearchProvider
from .serializers import UserCardSerializer

HEAD_RESULTS = 8
TAIL_RESULTS = 8

User = get_user_model()


class SearchUsers(SearchProvider):
    name = gettext_lazy("Users")
    icon = "people"
    url = "users"

    def allow_search(self):
        if not self.request.user_acl["can_search_users"]:
            raise PermissionDenied(_("You don't have permission to search users."))

    def search(self, query, page=1):
        if query:
            results = search_users(
                search_disabled=self.request.user.is_staff, username=query
            )
        else:
            results = []

        return {
            "results": UserCardSerializer(results, many=True).data,
            "count": len(results),
        }


def search_users(**filters):
    queryset = User.objects.order_by("slug").select_related(
        "rank", "ban_cache", "online_tracker"
    )

    if not filters.get("search_disabled", False):
        queryset = queryset.filter(is_active=True)

    username = filters.get("username").lower()

    results = []

    # lets grab head and tail results:
    results += list(queryset.filter(slug__startswith=username)[:HEAD_RESULTS])
    results += list(
        queryset.filter(slug__contains=username).exclude(
            pk__in=[r.pk for r in results]
        )[:TAIL_RESULTS]
    )

    return results



# Generated by Django 3.2.15 on 2022-12-29 19:55

from django.db import migrations


def create_index_message_setting(apps, _):
    Setting = apps.get_model("misago_conf", "Setting")
    Setting.objects.create(setting="index_message", is_public=False)


class Migration(migrations.Migration):

    dependencies = [
        ("misago_conf", "0005_add_sso_settings"),
    ]

    operations = [migrations.RunPython(create_index_message_setting)]



from .models import Theme


def get_active_theme():
    active_theme = Theme.objects.get(is_active=True)
    themes = active_theme.get_ancestors(include_self=True)
    themes = themes.prefetch_related("css")

    include_defaults = False
    styles = []

    for theme in themes:
        if theme.is_default:
            include_defaults = True
        for css in theme.css.all():
            if css.url:
                styles.append(css.url)
            elif css.source_needs_building:
                if css.build_file:
                    styles.append(css.build_file.url)
            else:
                styles.append(css.source_file.url)

    return {"include_defaults": include_defaults, "styles": styles}



from ..permissions import can_edit_profile_details


def serialize_profilefields_data(request, profilefields, user):
    data = {"id": user.pk, "groups": [], "edit": False}

    can_edit = can_edit_profile_details(request.user_acl, user)
    has_editable_fields = False

    for group in profilefields.get_fields_groups():
        group_fields = []
        for field in group["fields"]:
            display_data = field.get_display_data(request, user)
            if display_data:
                group_fields.append(display_data)
            if can_edit and field.is_editable(request, user):
                has_editable_fields = True
        if group_fields:
            data["groups"].append({"name": group["name"], "fields": group_fields})

    data["edit"] = can_edit and has_editable_fields

    return data



from unittest.mock import Mock

import pytest
from django.utils.functional import SimpleLazyObject

from ..middleware import socialauth_providers_middleware


@pytest.fixture
def get_response():
    return Mock()


class PlainRequest:
    pass


@pytest.fixture
def plain_request():
    return PlainRequest()


def test_middleware_sets_attr_on_request(db, get_response, plain_request):
    middleware = socialauth_providers_middleware(get_response)
    middleware(plain_request)
    assert hasattr(plain_request, "socialauth")


def test_attr_set_by_middleware_on_request_is_lazy_object(
    db, get_response, plain_request
):
    middleware = socialauth_providers_middleware(get_response)
    middleware(plain_request)
    assert isinstance(plain_request.socialauth, SimpleLazyObject)


def test_middleware_calls_get_response(db, get_response, plain_request):
    middleware = socialauth_providers_middleware(get_response)
    middleware(plain_request)
    get_response.assert_called_once()


def test_middleware_is_not_reading_from_db(
    db, get_response, plain_request, django_assert_num_queries
):
    with django_assert_num_queries(0):
        middleware = socialauth_providers_middleware(get_response)
        middleware(plain_request)


def test_middleware_is_not_reading_from_cache(db, mocker, get_response, plain_request):
    cache_get = mocker.patch("django.core.cache.cache.get")
    middleware = socialauth_providers_middleware(get_response)
    middleware(plain_request)
    cache_get.assert_not_called()



import random
import time

from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand

from ....core.management.progressbar import show_progress

User = get_user_model()


class Command(BaseCommand):
    help = "Adds random followers for testing purposes"

    def handle(self, *args, **options):
        total_users = User.objects.count()
        if total_users < 2:
            self.stderr.write(
                "At least two users must exist in the databse in order for "
                "fake followers creation to be possible.\n"
            )
            return

        message = "Adding fake followers to %s users...\n"
        self.stdout.write(message % total_users)

        total_followers = 0
        processed_count = 0

        start_time = time.time()
        show_progress(self, processed_count, total_users)
        for user in User.objects.iterator():
            user.followed_by.clear()
            followers_to_create = random.randint(0, total_users - 1)
            while followers_to_create:
                # There's 34% chance we'll skip follower creation
                if random.randint(0, 100) > 34:
                    new_follower = (
                        User.objects.exclude(pk=user.pk).order_by("?")[:1].first()
                    )
                    if not user.is_following(new_follower):
                        user.follows.add(new_follower)
                followers_to_create -= 1

            processed_count += 1
            show_progress(self, processed_count, total_users)

        self.stdout.write("\nSynchronizing users...")
        for user in User.objects.iterator():
            user.followers = user.followed_by.count()
            user.following = user.follows.count()
            user.save(update_fields=["followers", "following"])

        total_time = time.time() - start_time
        total_humanized = time.strftime("%H:%M:%S", time.gmtime(total_time))
        message = "\nSuccessfully added %s fake followers in %s"
        self.stdout.write(message % (total_followers, total_humanized))



from django.core.exceptions import PermissionDenied
from django.utils.translation import gettext as _
from rest_framework.response import Response

from ....acl.objectacl import add_acl_to_obj
from ....categories import THREADS_ROOT_NAME
from ....categories.models import Category
from ...permissions import can_start_thread
from ...threadtypes import trees_map


def thread_start_editor(request):
    if request.user.is_anonymous:
        raise PermissionDenied(_("You need to be signed in to start threads."))

    # list of categories that allow or contain subcategories that allow new threads
    available = []
    categories = []

    queryset = Category.objects.filter(
        pk__in=request.user_acl["browseable_categories"],
        tree_id=trees_map.get_tree_id_for_root(THREADS_ROOT_NAME),
    ).order_by("-lft")

    for category in queryset:
        add_acl_to_obj(request.user_acl, category)

        post = False
        if can_start_thread(request.user_acl, category):
            post = {
                "close": bool(category.acl["can_close_threads"]),
                "hide": bool(category.acl["can_hide_threads"]),
                "pin": category.acl["can_pin_threads"],
            }

            available.append(category.pk)
            available.append(category.parent_id)
        elif category.pk in available:
            available.append(category.parent_id)

        categories.append(
            {
                "id": category.pk,
                "name": category.name,
                "level": category.level - 1,
                "post": post,
            }
        )

    # list only categories that allow new threads,
    # or contains subcategory that allows one
    cleaned_categories = []
    for category in reversed(categories):
        if category["id"] in available:
            cleaned_categories.append(category)

    if not cleaned_categories:
        raise PermissionDenied(
            _(
                "No categories that allow new threads "
                "are available to you at the moment."
            )
        )

    return Response(cleaned_categories)



from datetime import timedelta
from unittest.mock import Mock

import pytest
from django.contrib.auth import get_user_model
from django.test import override_settings
from django.urls import reverse
from django.utils import timezone

from ...conf.test import override_dynamic_settings
from ...test import assert_contains
from ...users.datadownloads import request_user_data_download
from ...users.models import DataDownload
from ...users.test import create_test_user
from ..views.index import (
    check_cache,
    check_data_downloads,
    check_debug_status,
    check_forum_address,
    check_https,
    check_inactive_users,
)

User = get_user_model()

admin_link = reverse("misago:admin:index")


def test_cache_check_passess_if_cache_can_be_read_after_setting(mocker):
    mocker.patch("misago.admin.views.index.cache.set")
    mocker.patch("misago.admin.views.index.cache.get", return_value="ok")
    assert check_cache() == {"is_ok": True}


def test_cache_check_passess_if_cache_cant_be_read_after_setting(mocker):
    mocker.patch("misago.admin.views.index.cache.set")
    mocker.patch("misago.admin.views.index.cache.get", return_value=None)
    assert check_cache() == {"is_ok": False}


def test_warning_about_disabled_cache_is_displayed_on_checks_list(mocker, admin_client):
    mocker.patch("misago.admin.views.index.check_cache", return_value={"is_ok": False})
    response = admin_client.get(admin_link)
    assert_contains(response, "Cache is disabled")


def test_data_downloads_processing_check_passes_if_there_are_no_data_downloads(db):
    assert check_data_downloads() == {"is_ok": True, "count": 0}


def test_ready_data_downloads_are_excluded_by_check(user):
    data_download = request_user_data_download(user)
    data_download.status = DataDownload.STATUS_READY
    data_download.save()

    assert check_data_downloads() == {"is_ok": True, "count": 0}


def test_expired_data_downloads_are_excluded_by_check(user):
    data_download = request_user_data_download(user)
    data_download.status = DataDownload.STATUS_EXPIRED
    data_download.save()

    assert check_data_downloads() == {"is_ok": True, "count": 0}


def test_recent_pending_data_downloads_are_excluded_by_check(user):
    data_download = request_user_data_download(user)
    data_download.status = DataDownload.STATUS_PENDING
    data_download.save()

    assert check_data_downloads() == {"is_ok": True, "count": 0}


def test_recent_processing_data_downloads_are_excluded_by_check(user):
    data_download = request_user_data_download(user)
    data_download.status = DataDownload.STATUS_PROCESSING
    data_download.save()

    assert check_data_downloads() == {"is_ok": True, "count": 0}


def test_check_fails_if_there_are_old_processing_data_downloads(user):
    data_download = request_user_data_download(user)
    data_download.status = DataDownload.STATUS_PROCESSING
    data_download.requested_on = timezone.now() - timedelta(days=5)
    data_download.save()

    assert check_data_downloads() == {"is_ok": False, "count": 1}


def test_check_fails_if_there_are_old_pending_data_downloads(user):
    data_download = request_user_data_download(user)
    data_download.status = DataDownload.STATUS_PENDING
    data_download.requested_on = timezone.now() - timedelta(days=5)
    data_download.save()

    assert check_data_downloads() == {"is_ok": False, "count": 1}


def test_warning_about_unprocessed_data_downloads_is_displayed_on_checks_list(
    admin_client, user
):
    data_download = request_user_data_download(user)
    data_download.status = DataDownload.STATUS_PENDING
    data_download.requested_on = timezone.now() - timedelta(days=5)
    data_download.save()

    response = admin_client.get(admin_link)
    assert_contains(response, "There is 1 unprocessed data download request")


class RequestMock:
    absolute_uri = "https://misago-project.org/somewhere/"

    def __init__(self, settings):
        self.settings = settings

    def build_absolute_uri(self, location):
        assert location == "/"
        return self.absolute_uri


@pytest.fixture
def request_mock(dynamic_settings):
    return RequestMock(dynamic_settings)


incorrect_address = "http://somewhere.com"
correct_address = RequestMock.absolute_uri


@override_dynamic_settings(forum_address=None)
def test_forum_address_check_handles_setting_not_configured(request_mock):
    result = check_forum_address(request_mock)
    assert result == {
        "is_ok": False,
        "set_address": None,
        "correct_address": request_mock.absolute_uri,
    }


@override_dynamic_settings(forum_address=incorrect_address)
def test_forum_address_check_detects_invalid_address_configuration(request_mock):
    result = check_forum_address(request_mock)
    assert result == {
        "is_ok": False,
        "set_address": incorrect_address,
        "correct_address": request_mock.absolute_uri,
    }


@override_dynamic_settings(forum_address=correct_address)
def test_forum_address_check_detects_valid_address_configuration(request_mock):
    result = check_forum_address(request_mock)
    assert result == {
        "is_ok": True,
        "set_address": correct_address,
        "correct_address": request_mock.absolute_uri,
    }


@override_dynamic_settings(forum_address=None)
def test_warning_about_unset_forum_address_is_displayed_on_checks_list(admin_client):
    response = admin_client.get(admin_link)
    assert_contains(response, "address")


@override_dynamic_settings(forum_address=incorrect_address)
def test_warning_about_incorrect_forum_address_is_displayed_on_checks_list(
    admin_client,
):
    response = admin_client.get(admin_link)
    assert_contains(response, "address")


@override_settings(DEBUG=False)
def test_debug_check_passess_if_debug_is_disabled():
    assert check_debug_status() == {"is_ok": True}


@override_settings(DEBUG=True)
def test_debug_check_fails_if_debug_is_enabled():
    assert check_debug_status() == {"is_ok": False}


@override_settings(DEBUG=True)
def test_warning_about_enabled_debug_is_displayed_on_checks_list(admin_client):
    response = admin_client.get(admin_link)
    assert_contains(response, "site is running in DEBUG mode")


def test_https_check_passes_if_site_is_accessed_over_https():
    request = Mock(is_secure=Mock(return_value=True))
    assert check_https(request) == {"is_ok": True}


def test_https_check_fails_if_site_is_accessed_without_https(mocker):
    request = Mock(is_secure=Mock(return_value=False))
    assert check_https(request) == {"is_ok": False}


def test_warning_about_accessing_site_without_https_is_displayed_on_checks_list(
    admin_client,
):
    response = admin_client.get(admin_link)
    assert_contains(response, "site is not running over HTTPS")


def test_inactive_users_check_passess_if_there_are_no_inactive_users(db):
    assert check_inactive_users() == {"is_ok": True, "count": 0}


def test_inactive_users_check_passess_if_there_are_less_than_eleven_inactive_users(db):
    for i in range(10):
        create_test_user(
            "User%s" % i,
            "user%s@example.com" % i,
            requires_activation=User.ACTIVATION_USER,
        )
    assert check_inactive_users() == {"is_ok": True, "count": 10}


def test_inactive_users_check_fails_if_there_are_more_than_ten_inactive_users(db):
    for i in range(11):
        create_test_user(
            "User%s" % i,
            "user%s@example.com" % i,
            requires_activation=User.ACTIVATION_USER,
        )
    assert check_inactive_users() == {"is_ok": False, "count": 11}


def test_warning_about_inactive_users_is_displayed_on_checks_list(admin_client):
    for i in range(11):
        create_test_user(
            "User%s" % i,
            "user%s@example.com" % i,
            requires_activation=User.ACTIVATION_USER,
        )

    response = admin_client.get(admin_link)
    assert_contains(response, "There are 11 inactive users accounts")



import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Setting",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("setting", models.CharField(unique=True, max_length=255)),
                ("name", models.CharField(max_length=255)),
                ("description", models.TextField(null=True, blank=True)),
                ("legend", models.CharField(max_length=255, null=True, blank=True)),
                ("order", models.IntegerField(default=0, db_index=True)),
                ("dry_value", models.TextField(null=True, blank=True)),
                ("default_value", models.TextField(null=True, blank=True)),
                ("python_type", models.CharField(default="string", max_length=255)),
                ("is_public", models.BooleanField(default=False)),
                ("is_lazy", models.BooleanField(default=False)),
                ("form_field", models.CharField(default="text", max_length=255)),
                ("field_extra", models.JSONField()),
            ],
            options={},
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="SettingsGroup",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("key", models.CharField(unique=True, max_length=255)),
                ("name", models.CharField(max_length=255)),
                ("description", models.TextField(null=True, blank=True)),
            ],
            options={},
            bases=(models.Model,),
        ),
        migrations.AddField(
            model_name="setting",
            name="group",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to="misago_conf.SettingsGroup",
                to_field="id",
            ),
            preserve_default=True,
        ),
    ]



from importlib import import_module

from ..conf import settings

_NOT_INITIALIZED_ERROR = (
    "PermissionProviders instance has to load providers with load() "
    "before get_obj_type_annotators(), get_user_acl_serializers(), "
    "list() or dict() methods will be available."
)

_ALREADY_INITIALIZED_ERROR = (
    "PermissionProviders instance has already loaded providers and "
    "acl_annotator or user_acl_serializer are no longer available."
)


class PermissionProviders:
    """manager for permission providers"""

    def __init__(self):
        self._initialized = False
        self._providers = []
        self._providers_dict = {}

        self._annotators = {}
        self._user_acl_serializers = []

    def load(self):
        if self._initialized:
            raise RuntimeError("providers are already loaded")

        self._register_providers()
        self._coerce_dict_values_to_tuples(self._annotators)
        self._user_acl_serializers = tuple(self._user_acl_serializers)
        self._initialized = True

    def _register_providers(self):
        for namespace in settings.MISAGO_ACL_EXTENSIONS:
            self._providers.append((namespace, import_module(namespace)))
            self._providers_dict[namespace] = import_module(namespace)

            if hasattr(self._providers_dict[namespace], "register_with"):
                self._providers_dict[namespace].register_with(self)

    def _coerce_dict_values_to_tuples(self, types_dict):
        for hashType in types_dict.keys():
            types_dict[hashType] = tuple(types_dict[hashType])

    def acl_annotator(self, hashable_type, func):
        """registers ACL annotator for specified types"""
        assert not self._initialized, _ALREADY_INITIALIZED_ERROR
        self._annotators.setdefault(hashable_type, []).append(func)

    def user_acl_serializer(self, func):
        """registers ACL serializer for specified types"""
        assert not self._initialized, _ALREADY_INITIALIZED_ERROR
        self._user_acl_serializers.append(func)

    def get_obj_type_annotators(self, obj):
        assert self._initialized, _NOT_INITIALIZED_ERROR
        return self._annotators.get(obj.__class__, [])

    def get_user_acl_serializers(self):
        assert self._initialized, _NOT_INITIALIZED_ERROR
        return self._user_acl_serializers

    def list(self):
        assert self._initialized, _NOT_INITIALIZED_ERROR
        return self._providers

    def dict(self):
        assert self._initialized, _NOT_INITIALIZED_ERROR
        return self._providers_dict


providers = PermissionProviders()



import re

from django.forms import ValidationError
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _

from . import basefields


class BioField(basefields.UrlifiedTextareaProfileField):
    fieldname = "bio"
    label = _("Bio")


class RealNameField(basefields.TextProfileField):
    fieldname = "real_name"
    label = _("Real name")


class LocationField(basefields.TextProfileField):
    fieldname = "location"
    label = _("Location")


class GenderField(basefields.ChoiceProfileField):
    fieldname = "gender"
    label = _("Gender")

    choices = (
        ("", _("Not specified")),
        ("secret", _("Not telling")),
        ("female", _("Female")),
        ("male", _("Male")),
    )


class WebsiteField(basefields.UrlProfileField):
    fieldname = "website"
    label = _("Website")
    help_text = _(
        "If you own website in the internet you wish to share on your profile "
        "you may enter its address here. Remember to for it to be valid http "
        'address starting with either "http://" or "https://".'
    )


class SkypeIdField(basefields.TextProfileField):
    fieldname = "skype"
    label = _("Skype ID")
    help_text = _(
        "Entering your Skype ID in this field may invite other users to "
        "contact you over the Skype instead of via private threads."
    )


class TwitterHandleField(basefields.TextProfileField):
    fieldname = "twitter"
    label = _("Twitter handle")

    def get_help_text(self, user):
        return _(
            "If you own Twitter account, here you may enter your Twitter handle for "
            'other users to find you. Starting your handle with "@" sign is optional. '
            'Either "@%(slug)s" or "%(slug)s" are valid values.'
        ) % {"slug": user.slug}

    def get_value_display_data(self, request, user, value):
        return {"text": "@%s" % value, "url": "https://twitter.com/%s" % value}

    def clean(self, request, user, data):
        data = data.lstrip("@")
        if data and not re.search("^[A-Za-z0-9_]+$", data):
            raise ValidationError(gettext("This is not a valid twitter handle."))
        return data


class JoinIpField(basefields.TextProfileField):
    fieldname = "join_ip"
    label = _("Join IP")
    readonly = True

    def get_value_display_data(self, request, user, value):
        if not request.user_acl.get("can_see_users_ips"):
            return None

        if not user.joined_from_ip:
            return None

        return {"text": user.joined_from_ip}



# Generated by Django 1.10.5 on 2017-02-19 16:39
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("misago_users", "0006_update_settings")]

    operations = [
        migrations.AlterField(
            model_name="user",
            name="limits_private_thread_invites_to",
            field=models.PositiveIntegerField(
                choices=[(0, "Everybody"), (1, "Users I follow"), (2, "Nobody")],
                default=0,
            ),
        ),
        migrations.AlterField(
            model_name="user",
            name="subscribe_to_replied_threads",
            field=models.PositiveIntegerField(
                choices=[(0, "No"), (1, "Notify"), (2, "Notify with e-mail")], default=0
            ),
        ),
        migrations.AlterField(
            model_name="user",
            name="subscribe_to_started_threads",
            field=models.PositiveIntegerField(
                choices=[(0, "No"), (1, "Notify"), (2, "Notify with e-mail")], default=0
            ),
        ),
    ]



import os

import pytest
from django.core.files.uploadedfile import SimpleUploadedFile

from ...models import Icon


@pytest.fixture
def apple_touch_icon(db, image):
    return Icon.objects.create(
        type=Icon.TYPE_APPLE_TOUCH_ICON,
        image=SimpleUploadedFile("image.png", image, "image/png"),
    )


def test_new_touch_icon_can_be_set(admin_client, admin_link, image):
    admin_client.post(
        admin_link,
        {"apple_touch_icon": SimpleUploadedFile("image.png", image, "image/png")},
    )
    Icon.objects.get(type=Icon.TYPE_APPLE_TOUCH_ICON)


def test_setting_new_touch_icon_removes_old_one(
    admin_client, admin_link, image, apple_touch_icon
):
    admin_client.post(
        admin_link,
        {"apple_touch_icon": SimpleUploadedFile("image.png", image, "image/png")},
    )

    with pytest.raises(Icon.DoesNotExist):
        apple_touch_icon.refresh_from_db()


def test_setting_new_touch_icon_removes_old_one_image_file(
    admin_client, admin_link, image, apple_touch_icon
):
    admin_client.post(
        admin_link,
        {"apple_touch_icon": SimpleUploadedFile("image.png", image, "image/png")},
    )

    assert not os.path.exists(apple_touch_icon.image.path)


def test_submitting_form_without_new_icon_does_not_delete_old_one(
    admin_client, admin_link, apple_touch_icon
):
    admin_client.post(admin_link, {})
    apple_touch_icon.refresh_from_db()


def test_icon_can_be_deleted_without_setting_new_one(
    admin_client, admin_link, apple_touch_icon
):
    admin_client.post(admin_link, {"apple_touch_icon_delete": "1"})

    with pytest.raises(Icon.DoesNotExist):
        apple_touch_icon.refresh_from_db()


def test_deleting_icon_also_deletes_its_image_file(
    admin_client, admin_link, apple_touch_icon
):
    admin_client.post(admin_link, {"apple_touch_icon_delete": "1"})
    assert not os.path.exists(apple_touch_icon.image.path)


def test_uploading_invalid_icon_does_not_remove_current_icon(
    admin_client, admin_link, apple_touch_icon, image_small
):
    admin_client.post(
        admin_link,
        {"apple_touch_icon": SimpleUploadedFile("image.png", image_small, "image/png")},
    )

    apple_touch_icon.refresh_from_db()


def test_icon_is_not_set_because_it_was_not_square(
    admin_client, admin_link, image_non_square
):
    admin_client.post(
        admin_link,
        {
            "apple_touch_icon": SimpleUploadedFile(
                "image.png", image_non_square, "image/png"
            )
        },
    )

    assert not Icon.objects.exists()


def test_icon_is_not_set_because_it_was_too_small(
    admin_client, admin_link, image_small
):
    admin_client.post(
        admin_link,
        {"apple_touch_icon": SimpleUploadedFile("image.png", image_small, "image/png")},
    )

    assert not Icon.objects.exists()



from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from . import ThreadType
from ...categories import PRIVATE_THREADS_ROOT_NAME


class PrivateThread(ThreadType):
    root_name = PRIVATE_THREADS_ROOT_NAME

    def get_category_name(self, category):
        return _("Private threads")

    def get_category_absolute_url(self, category):
        return reverse("misago:private-threads")

    def get_category_last_thread_url(self, category):
        return reverse(
            "misago:private-thread",
            kwargs={"slug": category.last_thread_slug, "pk": category.last_thread_id},
        )

    def get_category_last_thread_new_url(self, category):
        return reverse(
            "misago:private-thread-new",
            kwargs={"slug": category.last_thread_slug, "pk": category.last_thread_id},
        )

    def get_category_last_post_url(self, category):
        return reverse(
            "misago:private-thread-last",
            kwargs={"slug": category.last_thread_slug, "pk": category.last_thread_id},
        )

    def get_thread_absolute_url(self, thread, page=1):
        if page > 1:
            return reverse(
                "misago:private-thread",
                kwargs={"slug": thread.slug, "pk": thread.pk, "page": page},
            )

        return reverse(
            "misago:private-thread", kwargs={"slug": thread.slug, "pk": thread.pk}
        )

    def get_thread_last_post_url(self, thread):
        return reverse(
            "misago:private-thread-last", kwargs={"slug": thread.slug, "pk": thread.pk}
        )

    def get_thread_new_post_url(self, thread):
        return reverse(
            "misago:private-thread-new", kwargs={"slug": thread.slug, "pk": thread.pk}
        )

    def get_thread_api_url(self, thread):
        return reverse("misago:api:private-thread-detail", kwargs={"pk": thread.pk})

    def get_thread_editor_api_url(self, thread):
        return reverse(
            "misago:api:private-thread-post-editor", kwargs={"thread_pk": thread.pk}
        )

    def get_thread_posts_api_url(self, thread):
        return reverse(
            "misago:api:private-thread-post-list", kwargs={"thread_pk": thread.pk}
        )

    def get_post_merge_api_url(self, thread):
        return reverse(
            "misago:api:private-thread-post-merge", kwargs={"thread_pk": thread.pk}
        )

    def get_post_absolute_url(self, post):
        return reverse(
            "misago:private-thread-post",
            kwargs={"slug": post.thread.slug, "pk": post.thread.pk, "post": post.pk},
        )

    def get_post_api_url(self, post):
        return reverse(
            "misago:api:private-thread-post-detail",
            kwargs={"thread_pk": post.thread_id, "pk": post.pk},
        )

    def get_post_likes_api_url(self, post):
        return reverse(
            "misago:api:private-thread-post-likes",
            kwargs={"thread_pk": post.thread_id, "pk": post.pk},
        )

    def get_post_editor_api_url(self, post):
        return reverse(
            "misago:api:private-thread-post-editor",
            kwargs={"thread_pk": post.thread_id, "pk": post.pk},
        )

    def get_post_edits_api_url(self, post):
        return reverse(
            "misago:api:private-thread-post-edits",
            kwargs={"thread_pk": post.thread_id, "pk": post.pk},
        )

    def get_post_read_api_url(self, post):
        return reverse(
            "misago:api:private-thread-post-read",
            kwargs={"thread_pk": post.thread_id, "pk": post.pk},
        )



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots[
    "test_code_in_quote_bbcode_header_is_escaped 1"
] = """<aside class="quote-block">
<div class="quote-heading" data-noquote="1">@Us&quot;&gt;&lt;script&gt;alert(&quot;!&quot;)&lt;/script&gt;er</div>
<blockquote class="quote-body" data-author="@Us&quot;&gt;&lt;script&gt;alert(&quot;!&quot;)&lt;/script&gt;er" data-block="quote">
<p>Test</p>
</blockquote>
</aside>"""

snapshots[
    "test_code_in_quote_bbcode_is_escaped 1"
] = """<aside class="quote-block">
<div class="quote-heading" data-noquote="1"></div>
<blockquote class="quote-body" data-block="quote">
<p>&lt;script&gt;alert(&quot;!&quot;)&lt;/script&gt;</p>
</blockquote>
</aside>"""

snapshots[
    "test_code_in_quote_markdown_is_escaped 1"
] = """<blockquote>
<p>&lt;script&gt;alert(&quot;!&quot;)&lt;/script&gt;</p>
</blockquote>"""

snapshots[
    "test_inline_code_is_escaped 1"
] = "<p><code>&lt;script&gt;alert(&quot;!&quot;)&lt;/script&gt;</code></p>"

snapshots[
    "test_text_is_escaped 1"
] = "<p>&lt;script&gt;alert(&quot;!&quot;)&lt;/script&gt;</p>"



from rest_framework import serializers

from ..models import UsernameChange
from .user import UserSerializer as BaseUserSerializer

__all__ = ["UsernameChangeSerializer"]

UserSerializer = BaseUserSerializer.subset_fields("id", "username", "avatars", "url")


class UsernameChangeSerializer(serializers.ModelSerializer):
    user = UserSerializer(many=False, read_only=True)
    changed_by = UserSerializer(many=False, read_only=True)

    class Meta:
        model = UsernameChange
        fields = [
            "id",
            "user",
            "changed_by",
            "changed_by_username",
            "changed_on",
            "new_username",
            "old_username",
        ]



from django.utils.translation import gettext as _
from rest_framework.exceptions import ValidationError

from .models import Poll


class MergeConflictHandler:
    def __init__(self, threads):
        self.items = []
        self.choices = {0: None}

        self._is_valid = False
        self._resolution = None

        self.threads = threads
        self.populate_from_threads(threads)

        if len(self.items) == 1:
            self._is_valid = True
            self._resolution = self.items[0]

    def populate_from_threads(self, threads):
        raise NotImplementedError("merge handler must define populate_from_threads")

    def is_merge_conflict(self):
        return len(self.items) > 1

    def set_resolution(self, resolution):
        try:
            resolution_clean = int(resolution)
        except (TypeError, ValueError):
            return

        if resolution_clean in self.choices:
            self._resolution = self.choices[resolution_clean]
            self._is_valid = True

    def is_valid(self):
        return self._is_valid

    def get_resolution(self):
        return self._resolution or None


class BestAnswerMergeHandler(MergeConflictHandler):
    data_name = "best_answer"

    def populate_from_threads(self, threads):
        for thread in threads:
            if thread.has_best_answer:
                self.items.append(thread)
                self.choices[thread.pk] = thread
        self.items.sort(key=lambda thread: (thread.title, thread.id))

    def get_available_resolutions(self):
        resolutions = [[0, _("Unmark all best answers")]]
        for thread in self.items:
            resolutions.append([thread.pk, thread.title])
        return resolutions


class PollMergeHandler(MergeConflictHandler):
    data_name = "poll"

    def populate_from_threads(self, threads):
        for thread in threads:
            try:
                self.items.append(thread.poll)
                self.choices[thread.poll.id] = thread.poll
            except Poll.DoesNotExist:
                pass
        self.items.sort(key=lambda poll: poll.question)

    def get_available_resolutions(self):
        resolutions = [[0, _("Delete all polls")]]
        for poll in self.items:
            resolutions.append(
                [poll.id, "%s (%s)" % (poll.question, poll.thread.title)]
            )
        return resolutions


class MergeConflict:
    """
    Utility class single point of entry for detecting merge conflicts on different
    properties and validating user resolutions.
    """

    HANDLERS = (BestAnswerMergeHandler, PollMergeHandler)

    def __init__(self, data=None, threads=None):
        self.data = data or {}
        self._handlers = [Handler(threads) for Handler in self.HANDLERS]
        self._conflicts = [i for i in self._handlers if i.is_merge_conflict()]
        self.set_resolution(data)

    def is_merge_conflict(self):
        return bool(self._conflicts)

    def get_conflicting_fields(self):
        return [i.data_name for i in self._conflicts]

    def set_resolution(self, data):
        for handler in self._conflicts:
            data = self.data.get(handler.data_name)
            handler.set_resolution(data)

    def is_valid(self, raise_exception=False):
        if raise_exception:
            self.raise_exception()
        return all([i.is_valid() for i in self._handlers])

    def raise_exception(self):
        # if any choice was made by user, we are in validation stage of resolution
        for conflict in self._conflicts:
            if self.data.get(conflict.data_name) is not None:
                self.raise_validation_exception()
                break
        else:
            self.raise_resolutions_exception()

    def raise_validation_exception(self):
        errors = {}
        for conflict in self._conflicts:
            if not conflict.is_valid() or self.data.get(conflict.data_name) is None:
                errors[conflict.data_name] = [_("Invalid choice.")]
        if errors:
            raise ValidationError(errors)

    def raise_resolutions_exception(self):
        resolutions = {}
        for conflict in self._conflicts:
            key = "%ss" % conflict.data_name
            resolutions[key] = conflict.get_available_resolutions()
        if resolutions:
            raise ValidationError(resolutions)

    def get_resolution(self):
        resolved_handlers = [i for i in self._handlers if i.is_valid()]
        return {i.data_name: i.get_resolution() for i in resolved_handlers}



from datetime import timedelta

from django.utils import timezone

from ..conf import settings


def get_cutoff_date(settings, user=None):
    cutoff_date = timezone.now() - timedelta(days=settings.readtracker_cutoff)

    if user and user.is_authenticated and user.joined_on > cutoff_date:
        return user.joined_on
    return cutoff_date



from django.shortcuts import render
from django.urls import reverse

from .serializers import CategoryWithPosterSerializer as CategorySerializer
from .utils import get_categories_tree


def categories(request):
    categories_tree = get_categories_tree(request, join_posters=True)

    request.frontend_context.update(
        {
            "CATEGORIES": CategorySerializer(
                categories_tree, context=request, many=True
            ).data,
            "CATEGORIES_API": reverse("misago:api:category-list"),
        }
    )

    return render(
        request, "misago/categories/list.html", {"categories": categories_tree}
    )



from django.urls import reverse

from ...acl.test import patch_user_acl
from ..test import AuthenticatedUserTestCase, create_test_user


class UserDetailsApiTests(AuthenticatedUserTestCase):
    def test_api_has_no_showstoppers(self):
        """api outputs response for freshly created user"""
        response = self.client.get(
            reverse("misago:api:user-details", kwargs={"pk": self.user.pk})
        )

        self.assertEqual(response.status_code, 200)
        self.assertTrue(response.json()["edit"])

    def test_api_has_no_showstoppers_old_user(self):
        """api outputs response for freshly created user"""
        self.user.profile_fields = {
            "gender": "f",
            "bio": "Lorem ipsum dolor met, sit amet elit, si vis pacem bellum.",
        }
        self.user.save()

        response = self.client.get(
            reverse("misago:api:user-details", kwargs={"pk": self.user.pk})
        )

        self.assertEqual(response.status_code, 200)
        self.assertTrue(response.json()["edit"])

    def test_other_user(self):
        """api handles scenario when its other user looking at profile"""
        user = create_test_user("OtherUser", "otheruser@example.com")
        api_link = reverse("misago:api:user-details", kwargs={"pk": user.pk})

        # moderator has permission to edit details
        with patch_user_acl({"can_moderate_profile_details": True}):
            response = self.client.get(api_link)
            self.assertEqual(response.status_code, 200)
            self.assertTrue(response.json()["edit"])

        # non-moderator has no permission to edit details
        with patch_user_acl({"can_moderate_profile_details": False}):
            response = self.client.get(api_link)
            self.assertEqual(response.status_code, 200)
            self.assertFalse(response.json()["edit"])

    def test_nonexistant_user(self):
        """api handles nonexistant users"""
        api_link = reverse("misago:api:user-details", kwargs={"pk": self.user.pk + 123})

        response = self.client.get(api_link)
        self.assertEqual(response.status_code, 404)



from ..acl.objectacl import add_acl_to_obj
from ..readtracker import categoriestracker
from .models import Category


def get_categories_tree(
    request, parent=None, join_posters=False
):  # pylint: disable=too-many-branches
    if not request.user_acl["visible_categories"]:
        return []

    if parent:
        queryset = parent.get_descendants().order_by("lft")
    else:
        queryset = Category.objects.all_categories()

    queryset_with_acl = queryset.filter(id__in=request.user_acl["visible_categories"])
    if join_posters:
        queryset_with_acl = queryset_with_acl.select_related("last_poster")

    visible_categories = list(queryset_with_acl)

    categories_dict = {}
    categories_list = []

    parent_level = parent.level + 1 if parent else 1

    for category in visible_categories:
        category.subcategories = []
        categories_dict[category.pk] = category
        categories_list.append(category)

        if category.parent_id and category.level > parent_level:
            categories_dict[category.parent_id].subcategories.append(category)

    add_acl_to_obj(request.user_acl, categories_list)
    categoriestracker.make_read_aware(request, categories_list)

    for category in reversed(visible_categories):
        if category.acl["can_browse"]:
            category.parent = categories_dict.get(category.parent_id)
            if category.parent:
                category.parent.threads += category.threads
                category.parent.posts += category.posts

                if category.parent.last_post_on and category.last_post_on:
                    parent_last_post = category.parent.last_post_on
                    category_last_post = category.last_post_on
                    update_last_thead = parent_last_post < category_last_post
                elif not category.parent.last_post_on and category.last_post_on:
                    update_last_thead = True
                else:
                    update_last_thead = False

                if update_last_thead:
                    category.parent.last_post_on = category.last_post_on
                    category.parent.last_thread_id = category.last_thread_id
                    category.parent.last_thread_title = category.last_thread_title
                    category.parent.last_thread_slug = category.last_thread_slug
                    category.parent.last_poster_name = category.last_poster_name
                    category.parent.last_poster_slug = category.last_poster_slug

                if not category.is_read:
                    category.parent.is_read = False

    flat_list = []
    for category in categories_list:
        if category.level == parent_level:
            flat_list.append(category)
    return flat_list


def get_category_path(category):
    if category.special_role:
        return [category]

    category_path = []
    while category and category.level > 0:
        category_path.append(category)
        category = category.parent
    return [f for f in reversed(category_path)]



import pytest

from ...models import SocialAuthProvider


@pytest.fixture
def provider(db):
    return SocialAuthProvider.objects.create(
        provider="facebook", is_active=True, order=0
    )


@pytest.fixture
def other_provider(db):
    return SocialAuthProvider.objects.create(provider="github", is_active=True, order=1)


@pytest.fixture
def disabled_provider(db):
    return SocialAuthProvider.objects.create(
        provider="google-oauth2", is_active=False, order=0
    )



import json

from django.urls import reverse

from .. import test
from ...categories import PRIVATE_THREADS_ROOT_NAME
from ...categories.models import Category
from ...conf.test import override_dynamic_settings
from ..models import Thread
from ..test import patch_category_acl
from ..threadtypes import trees_map
from .test_threads_api import ThreadsApiTestCase


class ThreadsBulkDeleteApiTests(ThreadsApiTestCase):
    def setUp(self):
        super().setUp()

        self.api_link = reverse("misago:api:thread-list")

        self.threads = [
            test.post_thread(category=self.category, poster=self.user),
            test.post_thread(category=self.category),
            test.post_thread(category=self.category, poster=self.user),
        ]

    def delete(self, url, data=None):
        return self.client.delete(
            url, json.dumps(data), content_type="application/json"
        )

    def test_delete_anonymous(self):
        """anonymous users can't bulk delete threads"""
        self.logout_user()

        response = self.delete(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "This action is not available to guests."}
        )

    @patch_category_acl({"can_hide_threads": 2, "can_hide_own_threads": 2})
    def test_delete_no_ids(self):
        """api requires ids to delete"""
        response = self.delete(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "You have to specify at least one thread to delete."},
        )

    @patch_category_acl({"can_hide_threads": 2, "can_hide_own_threads": 2})
    def test_validate_ids(self):
        response = self.delete(self.api_link, True)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": 'Expected a list of items but got type "bool".'}
        )

        response = self.delete(self.api_link, "abbss")
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": 'Expected a list of items but got type "str".'}
        )

        response = self.delete(self.api_link, [1, 2, 3, "a", "b", "x"])
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "One or more thread ids received were invalid."}
        )

    @override_dynamic_settings(threads_per_page=4)
    @patch_category_acl({"can_hide_threads": 2, "can_hide_own_threads": 2})
    def test_validate_ids_length(self):
        """api validates that ids are list of ints"""
        response = self.delete(self.api_link, list(range(5)))
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "No more than 4 threads can be deleted at a single time."},
        )

    @patch_category_acl({"can_hide_threads": 2, "can_hide_own_threads": 2})
    def test_validate_thread_visibility(self):
        """api valdiates if user can see deleted thread"""
        unapproved_thread = self.threads[1]

        unapproved_thread.is_unapproved = True
        unapproved_thread.save()

        threads_ids = [p.id for p in self.threads]

        response = self.delete(self.api_link, threads_ids)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "One or more threads to delete could not be found."},
        )

        # no thread was deleted
        for thread in self.threads:
            Thread.objects.get(pk=thread.pk)

    @patch_category_acl({"can_hide_threads": 0, "can_hide_own_threads": 2})
    def test_delete_other_user_thread_no_permission(self):
        """api valdiates if user can delete other users threads"""
        other_thread = self.threads[1]

        response = self.delete(self.api_link, [p.id for p in self.threads])
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            [
                {
                    "thread": {"id": other_thread.pk, "title": other_thread.title},
                    "error": "You can't delete other users theads in this category.",
                }
            ],
        )

        # no threads are removed on failed attempt
        for thread in self.threads:
            Thread.objects.get(pk=thread.pk)

    @patch_category_acl(
        {"can_hide_threads": 2, "can_hide_own_threads": 2, "can_close_threads": False}
    )
    def test_delete_thread_closed_category_no_permission(self):
        """api tests category's closed state"""
        self.category.is_closed = True
        self.category.save()

        response = self.delete(self.api_link, [p.id for p in self.threads])

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            [
                {
                    "thread": {"id": thread.pk, "title": thread.title},
                    "error": "This category is closed. You can't delete threads in it.",
                }
                for thread in sorted(self.threads, key=lambda i: i.pk)
            ],
        )

    @patch_category_acl(
        {"can_hide_threads": 2, "can_hide_own_threads": 2, "can_close_threads": False}
    )
    def test_delete_thread_closed_no_permission(self):
        """api tests thread's closed state"""
        closed_thread = self.threads[1]
        closed_thread.is_closed = True
        closed_thread.save()

        response = self.delete(self.api_link, [p.id for p in self.threads])

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            [
                {
                    "thread": {"id": closed_thread.pk, "title": closed_thread.title},
                    "error": "This thread is closed. You can't delete it.",
                }
            ],
        )

    @patch_category_acl({"can_hide_threads": 2, "can_hide_own_threads": 2})
    def test_delete_private_thread(self):
        """attempt to delete private thread fails"""
        private_thread = self.threads[0]

        private_thread.category = Category.objects.get(
            tree_id=trees_map.get_tree_id_for_root(PRIVATE_THREADS_ROOT_NAME)
        )
        private_thread.save()

        private_thread.threadparticipant_set.create(user=self.user, is_owner=True)

        threads_ids = [p.id for p in self.threads]

        response = self.delete(self.api_link, threads_ids)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "One or more threads to delete could not be found."},
        )

        Thread.objects.get(pk=private_thread.pk)
